import os

from gxl_ai_utils.utils import utils_file


def handle():
    """"""
    data_dir = "/home/work_nfs14/xlgeng/asr_data_raw/test_data_raw/Leaderboard/datasets"
    datanames = os.listdir(data_dir)
    utils_file.print_list(datanames)
    for name in datanames:
        temp_data_dir = os.path.join(data_dir, name)
        wav_dict_path = os.path.join(temp_data_dir, 'wav.scp')
        text_dict_path = os.path.join(temp_data_dir, 'trans.txt')
        if not (os.path.exists(wav_dict_path) and os.path.exists(text_dict_path)):
            utils_file.logging_print('找不到对应的文件：', wav_dict_path, text_dict_path)
            if os.path.exists(os.path.join(temp_data_dir, 'metadata.tsv')):
                do_tsv2kaldi(os.path.join(temp_data_dir, 'metadata.tsv'), temp_data_dir)
            continue
        continue
        wav_dict = utils_file.load_dict_from_scp(wav_dict_path)
        text_dict = utils_file.load_dict_from_scp(text_dict_path)
        new_wav_dict = {}
        for key, value in wav_dict.items():
            if not value.startswith(temp_data_dir):
                new_wav_dict[key] = os.path.join(temp_data_dir, value)
            else:
                new_wav_dict[key] = value
        utils_file.write_dict_to_scp(new_wav_dict, wav_dict_path)
        utils_file.write_dict_to_scp(text_dict, os.path.join(temp_data_dir, 'text'))


def do_tsv2kaldi(tsv_path, parent_dir):
    wav_scp_path = os.path.join(parent_dir, 'wav.scp')
    text_scp_path = os.path.join(parent_dir, 'text')
    # 读取tsv文件
    with open(tsv_path, 'r', encoding='utf-8') as tsv_file:
        lines = tsv_file.readlines()
    wav_dict = {}
    text_dict = {}
    # 遍历tsv文件的每一行
    for i, line in enumerate(lines):
        if i == 0:
            continue
        parts = line.strip().split('\t')
        # print(parts)
        id = parts[0]
        wav_path = parts[1]
        text = parts[-1]
        if not wav_path.startswith(parent_dir):
            wav_path = os.path.join(parent_dir, wav_path)
        wav_dict[id] = wav_path
        text_dict[id] = text
    utils_file.write_dict_to_scp(wav_dict, wav_scp_path)
    utils_file.write_dict_to_scp(text_dict, text_scp_path)


def handle_2():
    data_dir = "/home/work_nfs14/xlgeng/asr_data_raw/test_data_raw/Leaderboard/datasets"
    datanames = os.listdir(data_dir)
    for name in datanames:
        temp_data_dir = os.path.join(data_dir, name)
        text_path = os.path.join(temp_data_dir, 'text')
        wav_path = os.path.join(temp_data_dir, 'wav.scp')
        data_list_path = os.path.join(temp_data_dir, 'data.list')
        if os.path.exists(text_path) and os.path.exists(wav_path):
            utils_file.do_convert_wav_text_scp_to_jsonl(wav_path, text_path, data_list_path)

def do_convert(parent_dir):
    temp_data_dir = parent_dir
    text_path = os.path.join(temp_data_dir, 'text')
    wav_path = os.path.join(temp_data_dir, 'wav.scp')
    data_list_path = os.path.join(temp_data_dir, 'data.list')
    if os.path.exists(text_path) and os.path.exists(wav_path):
        utils_file.do_convert_wav_text_scp_to_jsonl(wav_path, text_path, data_list_path)


if __name__ == '__main__':
    """"""
    do_convert('/home/work_nfs14/xlgeng/asr_data_raw/test_data_raw/Leaderboard/datasets/SPEECHIO_ASR_ZH00009')



from os import PathLike
from typing import Dict, List, Optional, Union
from wenet.paraformer.search import paraformer_beautify_result
from wenet.text.char_tokenizer import CharTokenizer
from wenet.text.tokenize_utils import tokenize_by_seg_dict


def read_seg_dict(path):
    seg_table = {}
    with open(path, 'r', encoding='utf8') as fin:
        for line in fin:
            arr = line.strip().split('\t')
            assert len(arr) == 2
            seg_table[arr[0]] = arr[1]
    return seg_table


class ParaformerTokenizer(CharTokenizer):

    def __init__(self,
                 symbol_table: Union[str, PathLike, Dict],
                 seg_dict: Optional[Union[str, PathLike, Dict]] = None,
                 split_with_space: bool = False,
                 connect_symbol: str = '',
                 unk='<unk>') -> None:
        super().__init__(symbol_table, None, split_with_space, connect_symbol,
                         unk)
        self.seg_dict = seg_dict
        if seg_dict is not None and not isinstance(seg_dict, Dict):
            self.seg_dict = read_seg_dict(seg_dict)

    def text2tokens(self, line: str) -> List[str]:
        assert self.seg_dict is not None

        # TODO(Mddct): duplicated here, refine later
        line = line.strip()
        if self.non_lang_syms_pattern is not None:
            parts = self.non_lang_syms_pattern.split(line)
            parts = [w for w in parts if len(w.strip()) > 0]
        else:
            parts = [line]

        tokens = []
        for part in parts:
            if part in self.non_lang_syms:
                tokens.append(part)
            else:
                tokens.extend(tokenize_by_seg_dict(self.seg_dict, part))
        return tokens

    def tokens2text(self, tokens: List[str]) -> str:
        return paraformer_beautify_result(tokens)



from gxl_ai_utils.utils import utils_file

from align import do_align
from mingmingshiti import do_fenci


def my_print(*args):
    print(*args, flush=True)


def get_respresent_str(lab_list, rec_list):
    str_list = []
    for i, item in enumerate(lab_list):
        if item == "[]":
            str_list.append(rec_list[i])
        else:
            str_list.append(item)
    str_i = "".join(str_list)
    return str_i


class Info:
    def __init__(self, key, wer, lab_str, rec_str):
        lab_str = lab_str.replace("lab: ", '')
        rec_str = rec_str.replace("rec: ", '')
        self.key = key
        self.wer = wer
        self.lab = lab_str
        self.rec = rec_str
        self.lab_list, self.rec_list = do_align(lab_str, rec_str)
        self.str_represent = get_respresent_str(self.lab_list, self.rec_list)
        self.words, self.words_index = do_fenci(self.str_represent)
        self.diff_dict = compare_lab_rec_diff(self)
        self.lab_without_blank = self.lab.replace(" ", "")

    def __str__(self):
        return f'{self.key} {self.wer} {self.lab} {self.rec}'

    def show_info(self):
        my_print("-" * 100)
        my_print('key: ', self.key)
        my_print('wer: ', self.wer)
        my_print('lab: ', self.lab)
        my_print('rec: ', self.rec)
        my_print('lab_list: ', self.lab_list)
        my_print('rec_list: ', self.rec_list)
        my_print('str_represent: ', self.str_represent)
        my_print('words: ', self.words)
        my_print('words_index: ', self.words_index)
        # my_print('diff_dict: ')
        # utils_file.print_dict(self.diff_dict)
        my_print("-" * 100)


def read_info_from_file(file_path):
    info_list = []
    with open(file_path, 'r') as file:
        lines = file.readlines()
        i = 0
        while i < len(lines):
            if lines[i].startswith("utt:"):
                key = lines[i].split()[1]
                wer = float(lines[i + 1].split()[1])
                lab_str = lines[i + 2].strip()
                rec_str = lines[i + 3].strip()
                info = Info(key, wer, lab_str, rec_str)
                info_list.append(info)
                i += 5
            else:
                i += 1
    return info_list


from mingmingshiti import extract_nouns


def read_data():
    """"""
    # for index, partition in enumerate(['common','short','long','repeat','messy','encourage']):
    #     print("handling {}".format(partition))
    #     for i in range(27):
    #         key = f'speechio_{i}'
    #         res = input(f"请输入speechio_{i}的{partition}的wer path\n:")
    #         if key not in res_dict:
    #             res_dict[key] = []
    #         res_dict[key].append(res)
    #         print(f"输入内容为：{res}，是否重新输入（Y/*）")
    #         if input() == 'Y':
    #             key = f'speechio_{i}'
    #             res = input(f"请输入speechio_{i}的{partition}的wer path\n:")
    #             print(f"输入内容为：{res}")
    #             if key not in res_dict:
    #                 res_dict[key] = []
    #             res_dict[key][index] = res
    # res_dict = utils_file.load_data_from_xlsx(output_path2,return_cols=False)
    # utils_file.write_dict_to_json(res_dict, output_path)

    # output_path = "./data/wer_paths.json"
    # output_path2 = "./data/wer_paths.xlsx"
    # res_dict = utils_file.load_dict_from_json(path=output_path)
    # res_dict2 = {}
    # output_dir = "./data/wer_paths_store/"
    # for key in res_dict:
    #     if key == "speechio":
    #         continue
    #     print(f'handle {key}')
    #     for index, path in enumerate(res_dict[key]):
    #         if key not in res_dict2:
    #             res_dict2[key] = []
    #         output_path_i = path.replace("/home/work_nfs8/xlgeng/new_workspace/wenet_gxl_salmonn/examples/aishell/salmonn/exp/salmonn_v15/5_epoch",output_dir)
    #         utils_file.makedir_for_file(output_path_i)
    #         utils_file.copy_file(path, f"{output_dir}/{key}_{index}.txt")
    #         res_dict2[key].append(output_path_i)
    # utils_file.write_dict_to_json(res_dict2, "./data/wer_path_d2s3.json")
    # utils_file.write_dict_to_xlsx(res_dict2, "./data/wer_paths2.xlsx", cols_pattern=False)

    # res_dict = utils_file.load_dict_from_json('./data/wer_path_d2s3.json')
    # for key,value in res_dict.items():
    #     if key == 'speechio':
    #         continue
    #     print(f'{key}:{value}')
    #     new_values= []
    #     for path_i in value:
    #         path_2 = path_i.replace("./data/wer_paths_store//","./data/wer_paths_store/")
    #         new_values.append(path_2)
    #     res_dict[key] = new_values
    # utils_file.write_dict_to_json(res_dict, "./data/wer_path_d2s3.json")
    # utils_file.write_dict_to_xlsx(res_dict, "./data/wer_paths2.xlsx", cols_pattern=False)


def get_word_index_by_char_index(char_index, words_index_list):
    for i, word_list in enumerate(words_index_list):
        if char_index in word_list:
            return i
    return -1


def compare_lab_rec_diff(info: Info):
    res_dict = {
    }
    str_i = info.str_represent
    lab_list, rec_list = info.lab_list, info.rec_list
    if len(lab_list) != len(rec_list):
        print("lab_list和rec_list长度不一致")
        print(lab_list)
        print(rec_list)
        exit(0)
    words_list, words_index_list = do_fenci(str_i)
    for i in range(len(lab_list)):
        if lab_list[i] != rec_list[i]:
            word_index = get_word_index_by_char_index(i, words_index_list)
            word_involved = words_list[word_index]
            if word_involved in res_dict:
                continue
            word_index_involved = words_index_list[word_index]
            try:
                word_chars_lab = [lab_list[i] for i in word_index_involved]
                word_chars_rec = [rec_list[i] for i in word_index_involved]
            except:
                print(word_index_involved)
                info.show_info()
                continue
            word_lab = "".join(word_chars_lab)
            word_rec = "".join(word_chars_rec)
            if '[]' in word_chars_lab:
                tag = "insert"
            elif '[]' in word_chars_rec:
                tag = "delete"
            else:
                tag = "replace"
            res_dict[word_involved] = {
                "lab": word_lab,
                "rec": word_rec,
                "tag": tag,
            }
    return res_dict


def showing():
    info_list = read_info_from_file("./data/wer_paths_store/test_step_44268_speechio_short-prompt_2/speechio_6/wer")
    print(len(info_list))
    str_list = []
    for info in info_list:
        str_list.append(info.lab.replace(" ", ""))
        print(info.lab.replace(" ", ""))
    # for info in info_list:
    #     info.show_info()


from Gxl2DTable import Gxl2DTable

wer_path_info_path = "./data/wer_paths2.json"
dict_temp1 = {'speechio': ['common', 'short', 'long', 'repeat', 'messy', 'encourage']}
dict_temp2 = utils_file.load_dict_from_json(wer_path_info_path)
dict_temp1.update(dict_temp2)
wer_table = Gxl2DTable.load_from_json(dict_temp1)


def do_diff_analysis(info_list1, info_list2, res_dict):
    """"""
    my_res_dict = {}
    my_res_dict2 = {}
    # 得到keys
    keys = [item.key for item in info_list1]
    info_list1_dict = {item.key: item for item in info_list1}
    info_list2_dict = {item.key: item for item in info_list2}
    total_err_word = 0
    for key in utils_file.tqdm(keys, total=len(keys), desc="analysising", disable=False):
        """"""
        if not key in info_list2_dict:
            utils_file.logging_print(f'{key} not in info_list2_dict, skipping')
            continue
        info1 = info_list1_dict[key]
        info2 = info_list2_dict[key]
        diff_dict1 = info1.diff_dict
        diff_dict2 = info2.diff_dict
        if len(diff_dict1) == 0 and len(diff_dict2) == 0:
            continue
        err_word_in_1 = list(diff_dict1.keys())
        err_word_in_2 = list(diff_dict2.keys())
        for err_word in err_word_in_1:
            if err_word in res_dict:
                total_err_word += 1
                if err_word in err_word_in_2:
                    err_item = {"key": key, "right": err_word, 'err_left': diff_dict1[err_word]['rec'],
                                'err_right': diff_dict2[err_word]['rec'], 'lab': info1.lab_without_blank}
                    my_res_dict2[err_word] = {
                        "frequency": my_res_dict2[err_word]["frequency"] + 1 if err_word in my_res_dict2 else 1,
                        "errs": my_res_dict2[err_word]["errs"] + [err_item] if err_word in my_res_dict2 else [err_item],
                    }
                else:
                    err_item = {"key": key, "right": err_word, 'err_left': diff_dict1[err_word]['rec'],
                                'err_right': err_word, 'lab': info1.lab_without_blank}
                    my_res_dict[err_word] = {
                        "frequency": my_res_dict[err_word]["frequency"] + 1 if err_word in my_res_dict else 1,
                        "errs": my_res_dict[err_word]["errs"] + [err_item] if err_word in my_res_dict else [err_item],
                    }
    return my_res_dict, my_res_dict2, total_err_word


def analysis_for_one_speechio(input_num: int):
    """"""
    desc_path = 'data/desc.scp'
    max_list_length = -1
    desc_dict = utils_file.load_dict_from_scp(desc_path)
    big_res_dict = {}  # col pattern
    for input_num in range(0, 27):
        utils_file.logging_print(f'start analysis speechio_{input_num}')
        dataset_name = f'speechio_{input_num}'
        big_res_dict[dataset_name] = ["short"]
        big_res_dict[desc_dict[dataset_name]] = ["long"]
        big_res_dict[f'{dataset_name}.'] = ["repeat"]
        big_res_dict[f'{dataset_name}..'] = ["messy"]
        big_res_dict[f'{dataset_name}...'] = ["encourage"]

        # 首先得到common的info_list
        wer_path = wer_table.get_value_by_row_index_col_key(input_num, 'common')
        info_list = read_info_from_file(wer_path)
        lab_without_blank_list = [item.lab_without_blank for item in info_list]
        # 得到命名实体的集合
        res_set, res_dict = extract_nouns(lab_without_blank_list)
        sorted_items = sorted(res_dict.items(), key=lambda item: item[1], reverse=True)
        sorted_items = sorted_items[:400]
        res_dict = {item[0]: item[1] for item in sorted_items if len(item[0]) > 1}

        # 然后得到short对应的info_list ,和common进行对比分析
        for partition in ['short', 'long', 'repeat', 'messy', 'encourage']:
            wer_path = wer_table.get_value_by_row_index_col_key(input_num, partition)
            info_list_short = read_info_from_file(wer_path)
            the_res_dict, the_res_dict2, total = do_diff_analysis(info_list, info_list_short, res_dict)
            num_res_dict = sum([item['frequency'] for item in the_res_dict.values()])
            num_res_dict2 = sum([item['frequency'] for item in the_res_dict2.values()])
            err_words_right = the_res_dict.keys()
            print(err_words_right)
            err_words_right_dict = {}
            for err_word in err_words_right:
                err_words_right_dict[err_word] = the_res_dict[err_word]['errs'][0]['err_left']
            print(err_words_right_dict)

            err_words_error = the_res_dict2.keys()
            print(err_words_error)
            utils_file.logging_print(
                f'speechio_{input_num} {partition} right num:{num_res_dict}, error num:{num_res_dict2},total:{total}')
            if partition == 'short':
                the_key = dataset_name
            elif partition == 'long':
                the_key = desc_dict[dataset_name]
            elif partition == 'repeat':
                the_key = f'{dataset_name}.'
            elif partition == 'messy':
                the_key = f'{dataset_name}..'
            elif partition == 'encourage':
                the_key = f'{dataset_name}...'
            else:
                the_key = dataset_name
            list_i = big_res_dict[the_key]
            list_i.append(f'提升词汇的次数:{num_res_dict}')
            list_i.append(f'未提升词的次数:{num_res_dict2}')
            list_i.append(f'提升比例:{num_res_dict / (num_res_dict + num_res_dict2)}')
            list_i.append(f'正确词汇->common不正确的词汇:')
            for key, value in err_words_right_dict.items():
                list_i.append(f'{key}->{value}')
            big_res_dict[the_key] = list_i
            if max_list_length < len(list_i):
                max_list_length = len(list_i)
        for key, value_list in big_res_dict.items():
            if len(value_list) < max_list_length:
                for i in range(0, max_list_length - len(value_list)):
                    value_list.append("")
        output_file_path = "data/final_res2.xlsx"
        utils_file.write_dict_to_xlsx(big_res_dict, output_file_path, cols_pattern=True)


def add_wer_info():
    """"""
    res_list1 = []
    res_list2 = []
    for input_num in range(0, 27):
        utils_file.logging_print(f'start analysis speechio_{input_num}')
        dataset_name = f'speechio_{input_num}'
        # 首先得到common的info_list
        common_wer_path = wer_table.get_value_by_row_index_col_key(input_num, 'common')
        wer_common, replace_common, delete_common, insert_common = utils_file.get_wer_all_from_wer_file(common_wer_path)
        print(wer_common, replace_common, delete_common, insert_common )
        # 然后得到short对应的info_list ,和common进行对比分析
        for partition in ['short', 'long', 'repeat', 'messy', 'encourage']:
            wer_path = wer_table.get_value_by_row_index_col_key(input_num, partition)
            wer_partition, replace_partition, delete_partition, insert_partition = utils_file.get_wer_all_from_wer_file(
                wer_path)
            print(wer_partition, replace_partition, delete_partition, insert_partition)
            res_list1.append(wer_partition / wer_common)
            temp_item = 'S:{},D:{},I:{}'.format(replace_partition / replace_common, delete_partition / delete_common,
                                                insert_partition / insert_common)
            res_list2.append(temp_item)
    res_dict = utils_file.load_data_from_xlsx('./data/final_res2 (1).xlsx')
    for i, key in enumerate(res_dict.keys()):
        value_list = res_dict[key]
        value_list.insert(1, res_list1[i])
        value_list.insert(2, res_list2[i])
        res_dict[key] = value_list

    output_file_path = "data/final_res2 (2).xlsx"
    utils_file.write_dict_to_xlsx(res_dict, output_file_path, cols_pattern=True)


if __name__ == '__main__':
    add_wer_info()



import argparse
import ast
import glob
import os
import sys
sys.path.insert(0,'../../../../')
import tqdm

from gxl_ai_utils.utils import utils_file

NEW2OLD_TABLE = {}
VAD_INFO_TABLE = {}
THREAD_NUM = 40 # 切割音频的进程
IF_START_BY_0 = True  # 切割后的音频的名字的序号,是否以0起始,还是以1起始.
VAD_TYPE = 1  # 0:文件,文件内多行,每行都是i:j .. 1:文件, 格式如下:[[i,j],[i,j],[i,j]..] .. 2: str, 格式如下:[[i,j],[i,j],[i,j]..] ... 3:vad str, 格式如下:i:j i:j i:j ...
TTS_FINAL_TYPE=1  # 0:简短list型.. 1:含txt的scp型
def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--tts_final_scp_name", type=str)
    parser.add_argument("--asr_final_scp_name", type=str)
    parser.add_argument("--do_split", type=bool, default=False)
    args = parser.parse_args()
    return args
def handle_path(input_new_tts_final_path, only_name=False):
    """"""
    # 郭钊版：/root/path/282492425329_CPKwq_86_6299.wav
    if not only_name:
        file_name = utils_file.get_file_pure_name_from_path(input_new_tts_final_path)
    else:
        file_name = input_new_tts_final_path
    file_name_item_list = file_name.strip().split("_")
    parent_new_name = file_name_item_list[0] + "_" + file_name_item_list[1]
    if file_name_item_list[2].startswith("VAD"):
        index = int(file_name_item_list[2].split("VAD")[1])
    else:
        index = int(file_name_item_list[2])
    millisecond_num = int(file_name_item_list[3])
    return parent_new_name, index, millisecond_num
def get_little_final(input_final_scp_path, output_dir):
    """从非常大的tts_final_scp中取出一小点，组成一个测试用的小的tts_final_scp文件"""
    all_dict = utils_file.load_dict_from_scp(input_final_scp_path)
    little_dict = utils_file.get_random_subdict(all_dict, 25)
    utils_file.write_dict_to_scp(little_dict, os.path.join(output_dir, f"final_scp_little.scp"))
def pre_do(final_tts_scp_path: str, output_dir: str):
    utils_file.logging_print('首先切分tts_final.scp,切分为6个')
    # final_tts_dict = utils_file.load_dict_from_scp(final_tts_scp_path)
    # tts_dict_list = utils_file.do_split_dict(final_tts_dict, 6)
    # for i, tts_dict in enumerate(tts_dict_list):
    #     i = i + 1
    #     utils_file.write_dict_to_scp(tts_dict, os.path.join(output_dir, f"tts_final_{i}.scp"))
    final_tts_list = utils_file.load_list_file_clean(final_tts_scp_path)
    list_list = utils_file.do_split_list(final_tts_list, 6)
    for i, list in enumerate(list_list):
        i = i + 1
        utils_file.write_list_to_file(list, os.path.join(output_dir, f"tts_final_{i}.scp"))
def do_handle_wav(long_wav_new_name, output_dir="./"):
    """
    处理一个长音频. 传入该长音频的new_name, 然后对其进行clean和切割.
    :param long_wav_new_name:
    :param output_dir:
    :return:
    """
    global NEW2OLD_TABLE, VAD_INFO_TABLE, THREAD_NUM
    the_old_wav_path = NEW2OLD_TABLE.get(long_wav_new_name, None)
    if the_old_wav_path is None:
        utils_file.logging_print(
            f"do_handle_wav(): old2new字典中不包含如下new_name的映射, long_wav_new_name={long_wav_new_name}"
        )
        return False
    if not os.path.exists(the_old_wav_path):
        utils_file.logging_print(
            f"do_handle_wav(): the_old_wav_path is not exist, long_wav_new_name:{long_wav_new_name},the_old_wav_path:{the_old_wav_path}"
        )
        return False
    long_old_clean_wav_dir = os.path.join(output_dir, "temp_long_old_clean_wav")
    utils_file.makedir_sil(long_old_clean_wav_dir)
    long_clean_path = os.path.join(long_old_clean_wav_dir, long_wav_new_name + ".wav")
    if not os.path.exists(long_clean_path):
        # 先规范化音频
        utils_file.do_clean_wav(the_old_wav_path, long_clean_path)
        # 开始切割音频
        vad_info4long_new_name = VAD_INFO_TABLE.get(long_wav_new_name, None)
        if vad_info4long_new_name is None:
            utils_file.logging_print(
                f"do_handle_wav(): vad_table不包含如下new_name的映射, long_wav_new_name={long_wav_new_name}"
            )
            return False
        final_wav_dir = os.path.join(output_dir, "final_wav")
        utils_file.makedir_sil(final_wav_dir)
        slicing_wav(long_clean_path, vad_info4long_new_name, final_wav_dir, long_wav_new_name, THREAD_NUM)
    else:
        utils_file.logging_print(
            f"do_handle_wav(): long_old_clean_path is exist, long_wav_new_name={long_wav_new_name}, 跳过该长音频的处理"
        )
    return True
def slicing_wav(input_file_path, vad_info4long_new_name, output_dir, wav_new_name, thread_num=40):
    """
    将一个音频,依据时间戳, 切分成若干小音频. 此处可以多线程并行处理
    """
    global IF_START_BY_0, VAD_TYPE
    sorted_list = get_vad_info_list(vad_info4long_new_name, VAD_TYPE)
    thread_num = thread_num if len(sorted_list) > thread_num else len(sorted_list)
    runner = utils_file.GxlFixedThreadPool(thread_num)
    for i, vad_info in enumerate(sorted_list):
        if IF_START_BY_0:
            i = i
        else:
            i = i + 1
        runner.add_thread(little_func4slicing_wav, [i, vad_info, input_file_path, output_dir, wav_new_name])
    runner.start()
def little_func4slicing_wav(i, vad_info, input_file_path, output_dir, wav_new_name):
    """

    :param i:
    :param vad_info:  [i,j], 毫秒数
    :param input_file_path:
    :param output_dir:
    :param wav_new_name:
    :return:
    """
    start_time = vad_info[0]
    end_time = vad_info[1]
    duration = int(end_time) - int(start_time)
    start_sample = int(start_time) * 16
    end_sample = int(end_time) * 16
    output_path = os.path.join(output_dir, f"{wav_new_name}_{i}_{duration}.wav")
    utils_file.do_extract_audio_segment(input_file_path, output_path, start_sample, end_sample)
def get_vad_info_list(vad_info4long_new_name, vad_type=0):
    """
    处理vad_info4long_new_name, 得到排过序的一个双层列表 [[s,e],[s,e],[s,e]...]
    :param vad_info_txt_path:
    :return:
    """
    sorted_list = []
    if vad_type == 0:
        """
        vad文件, 格式如下:
        i,j
        i,j
        i,j     
        """
        vad_info_str_list = utils_file.load_list_file_clean(vad_info4long_new_name)
        vad_info_list = []
        for vad_info_str in vad_info_str_list:
            vad_i = vad_info_str.strip().split(",")
            vad_info_list.append(vad_i)
        sorted_list = sorted(vad_info_list, key=lambda x: int(x[0]))
    elif vad_type == 1:
        """
        vad 文件, 格式如下:
        [[i,j],[i,j],[i,j]..]
        """
        str_one_line = utils_file.load_first_row_clean(vad_info4long_new_name)
        vad_info_list = ast.literal_eval(str_one_line)
        sorted_list = sorted(vad_info_list, key=lambda x: int(x[0]))
    elif vad_type == 2:
        """
        vad str, 格式如下:
        [[i,j],[i,j],[i,j]..]
        """
        # str_one_line = utils_file.load_first_row_clean(vad_info4long_new_name)
        vad_info_list = ast.literal_eval(vad_info4long_new_name)
        sorted_list = sorted(vad_info_list, key=lambda x: int(x[0]))
    elif vad_type == 3:
        """
        vad str, 格式如下:
        i:j i:j i:j ...
        """
        split_list = vad_info4long_new_name.strip().split()
        vad_info_list = []
        for res in split_list:
            res_i = res.split(":")[0]
            res_j = res.split(":")[1]
            vad_info_list.append([res_i, res_j])
        sorted_list = sorted(vad_info_list, key=lambda x: int(x[0]))
    return sorted_list
def do_get_vad_scp_file(input_dir, output_dir='./'):
    """
    tts处理流程中会对每一个音频生成一个vad_info的txt文件, 我们得到一个key vad_info.txt的字典
    :param input_dir:
    :return:
    """
    if os.path.isdir(input_dir):
        vad_res_dict = utils_file.get_scp_for_wav_dir(input_dir, suffix="txt")
    else:
        res_list = utils_file.load_list_file_clean(input_dir)
        vad_res_dict = {}
        for res_line in res_list:
            wav_path = res_line.split('|')[0]
            key = utils_file.get_file_pure_name_from_path(wav_path)
            info = res_line.split('|')[1].strip()
            vad_res_dict[key] = info
    little_res_dict = utils_file.get_subdict(vad_res_dict, 0, 10)
    utils_file.logging_print("do_get_vad_scp_file(): 小字典示例:")
    utils_file.print_dict(little_res_dict)
    utils_file.write_dict_to_scp(vad_res_dict, os.path.join(output_dir, "vad_res.scp"))

def do_get_old2new_scp_file(input_dir, output_dir='./'):
    """
    这个input_dir中包含大量的old2new_*.txt文件, 我们得到一个key old2new.txt的字典
    :param input_dir:
    :return:
    """
    if os.path.isfile(input_dir):
        utils_file.logging_print("do_get_old2new_scp_file(): input_dir是文件")
        utils_file.copy_file(input_dir, os.path.join(output_dir, "old2new.scp"))
        return
    old2new_path_list = glob.glob(f"{input_dir}/old2new*.scp")
    res_dict = {}
    for old2new_path in old2new_path_list:
        old2new_dict = utils_file.load_dict_from_scp(old2new_path)
        res_dict.update(old2new_dict)
    little_res_dict = utils_file.get_subdict(res_dict, 0, 10)
    utils_file.logging_print("do_get_old2new_scp_file(): 小字典示例:")
    utils_file.print_dict(little_res_dict)
    utils_file.write_dict_to_scp(res_dict, os.path.join(output_dir, "old2new.scp"))

def solution(final_tts_scp_path, output_dir, old2new_dir, input_vad_info_dir):
    utils_file.logging_print('首先从命令行中拿到要处理的tts_final.scp的子集')
    args = get_args()
    if args.do_split:
        utils_file.logging_print('开始do_split')
        pre_do(final_tts_scp_path, output_dir)
        utils_file.logging_print('do_split完成, 直接return')
        return
    tts_scp_path = os.path.join(output_dir, args.tts_final_scp_name)
    asr_final_path = os.path.join(output_dir, args.asr_final_scp_name)
    if os.path.exists(asr_final_path):
        utils_file.logging_print('asr_scp_path已经存在，直接删除')
        os.remove(asr_final_path)
    utils_file.logging_print('拿到成功\n')

    utils_file.logging_print('加载vad和old2new的信息,从而在output目录下得到vad_res.scp和old2new.scp,这一步基本是不用变动的,然后将dict信息加载到全局变量中')
    utils_file.makedir_sil(output_dir)
    do_get_vad_scp_file(input_vad_info_dir, output_dir)
    do_get_old2new_scp_file(old2new_dir, output_dir)
    NEW2OLD_TABLE.update(utils_file.load_dict_from_scp(os.path.join(output_dir, "old2new.scp")))
    VAD_INFO_TABLE.update(utils_file.load_dict_from_scp(os.path.join(output_dir, "vad_res.scp")))
    utils_file.logging_print('加载vad和old2new的信息完成\n')

    utils_file.logging_print('获取tts_scp对应的长音频的新名字列表')
    long_wav_new_names_list = []
    if TTS_FINAL_TYPE == 0:
        wav_path_list = utils_file.load_list_file_clean(tts_scp_path)
        for wav_path in tqdm.tqdm(wav_path_list, total=len(wav_path_list)):
            long_wav_new_name, _, _ = handle_path(wav_path, only_name=False)
            long_wav_new_names_list.append(long_wav_new_name)
    elif TTS_FINAL_TYPE == 1:
        tts_dict = utils_file.load_dict_from_scp(tts_scp_path)
        for key in tqdm.tqdm(tts_dict.keys(), total=len(tts_dict)):
            long_wav_new_name, _, _ = handle_path(key, only_name=True)
            long_wav_new_names_list.append(long_wav_new_name)
    long_wav_new_names_list = list(set(long_wav_new_names_list))
    utils_file.logging_print('获取tts_scp对应的长音频的新名字列表完成,其一共有:', len(long_wav_new_names_list), '\n')

    utils_file.logging_print('开始验证vad信息和old2new信息是否全面,从而得到最终真正能处理的长音频的新名字列表,以及他的数量')
    final_long_wav_new_names_list = []
    for long_wav_new_name in tqdm.tqdm(long_wav_new_names_list, total=len(long_wav_new_names_list)):
        if long_wav_new_name in VAD_INFO_TABLE and long_wav_new_name in NEW2OLD_TABLE:
            final_long_wav_new_names_list.append(long_wav_new_name)
        else:
            utils_file.logging_print(
                f"这个长音频 验证vad信息和old2new信息是否全面失败: long_wav_new_name={long_wav_new_name}"
            )
            if not (long_wav_new_name in VAD_INFO_TABLE):
                utils_file.logging_print(f"vad_info表中不包含该音频")
            if not (long_wav_new_name in NEW2OLD_TABLE):
                utils_file.logging_print(f"new2old表中不包含该音频")
    utils_file.logging_print('验证vad信息和old2new信息是否全面完成,其一共有:', len(final_long_wav_new_names_list), '\n')
    long_wav_new_names_list = final_long_wav_new_names_list



    utils_file.logging_print("开始逐条长音频得处理")
    for long_wav_new_name in tqdm.tqdm(long_wav_new_names_list, total=len(long_wav_new_names_list)):
        do_handle_wav(long_wav_new_name, output_dir)
    utils_file.logging_print('完全切割完毕')
    return
    #  ------------------后处理-----------------------------

def post_process(output_dir,tts_final_path,shard_output_dir, text_dir=None):
    utils_file.logging_print('开始后处理')
    # 这里的tts final_path 只是处理后的wav_list了，没了text信息
    utils_file.logging_print('开始得到tts_wav_dict')
    if TTS_FINAL_TYPE == 0:
        tts_wav_path_list = utils_file.load_list_file_clean(tts_final_path)
        tts_dict = {}
        for tts_wav_path in tts_wav_path_list:
            key = utils_file.get_file_pure_name_from_path(tts_wav_path)
            tts_dict[key] = tts_wav_path
        little_dict = utils_file.get_subdict(tts_dict, 0, 10)
        utils_file.print_dict(little_dict)
    elif TTS_FINAL_TYPE == 1:
        tts_dict = utils_file.load_dict_from_scp(tts_final_path)
        little_dict = utils_file.get_subdict(tts_dict, 0, 10)
        utils_file.print_dict(little_dict)
    utils_file.logging_print('得到tts_wav_dict成功,总的tts_wav_dict数为：', len(tts_dict))

    utils_file.logging_print('开始得到总的切割音频的wav_dict：')
    if os.path.exists(os.path.join(output_dir, 'split_wav_big.scp')):
        split_wav_big_dict = utils_file.load_dict_from_scp(os.path.join(output_dir, 'split_wav_big.scp'))
    else:
        asr_wav_dir = os.path.join(output_dir, 'final_wav')
        split_wav_big_dict = utils_file.get_scp_for_wav_dir(asr_wav_dir)
        utils_file.write_dict_to_scp(split_wav_big_dict, os.path.join(output_dir, 'split_wav_big.scp'))
    little_split_dict = utils_file.get_subdict(split_wav_big_dict, 0, 10)
    utils_file.print_dict(little_split_dict)
    utils_file.logging_print('得到总的切割音频：完成,总的切割音频数为：', len(split_wav_big_dict))

    utils_file.logging_print('然后验证一下和tts wav的key关系')

    tmp_key = list(tts_dict.keys())[0]
    if 'VAD' in tmp_key:
        utils_file.logging_print('含有VAD的key风格,重新设计split_wav_dict的key')
        new_split_dict = {}
        for key, value in split_wav_big_dict.items():
            name, i, d = handle_path(key, only_name=True)
            duration = int(d/1000) + 1
            new_key = f"{name}_VAD{i}_{duration}"
            new_split_dict[new_key] = value
        little_dict = utils_file.get_subdict(new_split_dict, 0, 10)
        utils_file.print_dict(little_dict)
        split_wav_big_dict = new_split_dict

    num = 10
    for key, value in tts_dict.items():
        long_new_name, index, duration = handle_path(key, only_name=True)
        # key_new = f"{long_new_name}_{index}"
        if key in split_wav_big_dict:
            utils_file.logging_print('-----------')
            utils_file.logging_print(f'tts key：{key}, value：{value}')
            utils_file.logging_print(f'asr key：{key}, value：{split_wav_big_dict[key]}')
            utils_file.logging_print('-----------')
            utils_file.logging_print('\n')
            num -= 1
            if num < 0:
                break
    is_verify = num < 0
    utils_file.logging_print(f"验证一下tts wav和asr wav的key关系：验证完成,验证结果为:{is_verify}")

    utils_file.logging_print('如果验证成功，那么就计算覆盖率')
    if is_verify:
        num = 0
        for key, value in tts_dict.items():
            long_new_name, index, duration = handle_path(key, only_name=True)
            # key_new = f"{long_new_name}_{index}"
            if key in split_wav_big_dict:
                num += 1
        utils_file.logging_print(f'tts_final覆盖率为：{num / len(tts_dict)}')

    if not is_verify:
        return
    utils_file.logging_print('开始得到wav.scp')
    wav_scp_dict = {}
    for key, value in tts_dict.items():
        long_new_name, index, duration = handle_path(key, only_name=True)
        # key_new = f"{long_new_name}_{index}"
        if key in split_wav_big_dict:
            wav_scp_dict[key] = split_wav_big_dict[key]
    utils_file.write_dict_to_scp(wav_scp_dict, os.path.join(output_dir, 'wav.scp'))
    utils_file.logging_print('得到wav.scp：完成, 长度为：', len(wav_scp_dict))

    utils_file.logging_print('开始得到text.scp')
    text_scp_dict = {}
    if TTS_FINAL_TYPE == 0:
        temp_txt_file_dict = {}
        txt_file_list = glob.glob(str(os.path.join(text_dir, '*.txt')))
        for txt_file in tqdm.tqdm(txt_file_list, total=len(txt_file_list)):
            key = utils_file.get_file_pure_name_from_path(txt_file)
            temp_txt_file_dict[key] = txt_file
        for key in tqdm.tqdm(wav_scp_dict.keys(), total=len(wav_scp_dict)):
            if key in temp_txt_file_dict:
                txt_file = temp_txt_file_dict[key]
                line_i = utils_file.load_first_row_clean(txt_file)
                if len(line_i) > 0:
                    list_i = line_i.split()
                    if len(list_i) >= 2:
                        key_tmp = list_i[0]
                        value = ' '.join(list_i[1:])
                        key = key.strip()
                        value = value.strip()
                        if key == key_tmp:
                            text_scp_dict[key] = value
                        else:
                            utils_file.logging_print(f'error: key不相等，key：{key}, key_tmp：{key_tmp}, value:{value}')
                    else:
                        utils_file.logging_print(f'error: line_i没两个item，key：{key}, line_i：{line_i}')
    elif TTS_FINAL_TYPE == 1:
        for key, value in tqdm.tqdm(tts_dict.items()):
            temp_list = value.split()
            if len(temp_list) == 2:
                wav_path = temp_list[0]
                txt_path = temp_list[1]
                line_i = utils_file.load_first_row_clean(txt_path)
                if len(line_i) > 0:
                    list_i = line_i.split()
                    if len(list_i) >= 2:
                        key_tmp = list_i[0]
                        value = ' '.join(list_i[1:])
                        key = key.strip()
                        value = value.strip()
                        if key == key_tmp:
                            text_scp_dict[key] = value
                        else:
                            utils_file.logging_print(f'error: key不相等，key：{key}, key_tmp：{key_tmp}, value:{value}')
                    else:
                        utils_file.logging_print(f'error: line_i没两个item，key：{key}, line_i：{line_i}')
            else:
                utils_file.logging_print(f'error: value不是分为wav_path 和txt_path，key：{key}, value：{value}')
    utils_file.write_dict_to_scp(text_scp_dict, os.path.join(output_dir, 'text_with_punctuation'))
    utils_file.logging_print(f'text_scp得到完成,text_scp_dict长度为：{len(text_scp_dict)}')

    utils_file.logging_print('开始为text清除标点')
    text_with_punctuation_dict = utils_file.load_dict_from_scp(os.path.join(output_dir, 'text_with_punctuation'))
    text_no_punctuation_dict = {}
    for key, value in text_with_punctuation_dict.items():
        text_no_punctuation_dict[key] = utils_file.do_filter(value)
    utils_file.write_dict_to_scp(text_no_punctuation_dict, os.path.join(output_dir, 'text'))
    utils_file.logging_print('清除标点完成,len: ', len(text_no_punctuation_dict))

    utils_file.logging_print('开始打shard包')
    utils_file.do_make_shard_file(os.path.join(output_dir, "wav.scp"), os.path.join(output_dir, "text"),
                                  shard_output_dir, num_threads=50)
    utils_file.logging_print('打shard包完成')



if __name__ == '__main__':
    output_dir = "/home/node36_data/xlgeng/asr_data_from_pachong/gxl_output/lizhi_jiankang"
    utils_file.makedir_for_file_or_dir(output_dir)
    old2new_dir = "/home/work_nfs7/dkguo/lizhi/jiankang/init_lists"
    vad_info_dir = "/home/work_nfs7/dkguo/lizhi/jiankang/vad_result"
    # text_dir = "/home/work_nfs6/tlzuo/data_handle/lists/5txts"
    final_tts_scp_path = "/home/work_nfs14/zhguo/data/lizhi_jiankang/final_cmn.scp"
    # solution(final_tts_scp_path, output_dir, old2new_dir, vad_info_dir)
    sys.path.insert(0, '../')
    from gxl_common_utils.common_utils import post_process
    shard_output_dir = "/home/work_nfs9/xlgeng/asr_data_shard/pachong_data/4_lizhi_jiankang/"
    post_process(output_dir, final_tts_scp_path, shard_output_dir,TTS_FINAL_TYPE=TTS_FINAL_TYPE)


# Copyright (c) 2019 Shigeki Karita
#               2020 Mobvoi Inc (Binbin Zhang)
#               2022 Xingchen Song (sxc19@mails.tsinghua.edu.cn)
#               2022 58.com(Wuba) Inc AI Lab.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multi-Head Attention layer definition."""

import math
from typing import Tuple, Optional

import torch
from torch import nn
import torch.nn.functional as F
from wenet.transformer.attention import MultiHeadedAttention


class GroupedRelPositionMultiHeadedAttention(MultiHeadedAttention):
    """Multi-Head Attention layer with relative position encoding.
    Paper:
        https://arxiv.org/abs/1901.02860
        https://arxiv.org/abs/2109.01163
    Args:
        n_head (int): The number of heads.
        n_feat (int): The number of features.
        dropout_rate (float): Dropout rate.
    """

    def __init__(self, n_head, n_feat, dropout_rate, group_size=3):
        """Construct an RelPositionMultiHeadedAttention object."""
        super().__init__(n_head, n_feat, dropout_rate)
        # linear transformation for positional encoding
        self.linear_pos = nn.Linear(n_feat, n_feat, bias=False)
        self.group_size = group_size
        self.d_k = n_feat // n_head  # for GroupedAttention
        self.n_feat = n_feat
        # these two learnable bias are used in matrix c and matrix d
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        self.pos_bias_u = nn.Parameter(
            torch.Tensor(self.h, self.d_k * self.group_size))
        self.pos_bias_v = nn.Parameter(
            torch.Tensor(self.h, self.d_k * self.group_size))
        torch.nn.init.xavier_uniform_(self.pos_bias_u)
        torch.nn.init.xavier_uniform_(self.pos_bias_v)

    def rel_shift(self, x, zero_triu: bool = False):
        """Compute relative positinal encoding.
        Args:
            x (torch.Tensor): Input tensor (batch, time, size).
            zero_triu (bool): If true, return the lower triangular part of
                the matrix.
        Returns:
            torch.Tensor: Output tensor.
        """

        zero_pad = torch.zeros((x.size()[0], x.size()[1], x.size()[2], 1),
                               device=x.device,
                               dtype=x.dtype)
        x_padded = torch.cat([zero_pad, x], dim=-1)

        x_padded = x_padded.view(x.size()[0],
                                 x.size()[1],
                                 x.size(3) + 1, x.size(2))
        x = x_padded[:, :, 1:].view_as(x)

        if zero_triu:
            ones = torch.ones((x.size(2), x.size(3)))
            x = x * torch.tril(ones, x.size(3) - x.size(2))[None, None, :, :]

        return x

    def pad4group(self, Q, K, V, P, mask, group_size: int = 3):
        """
        q: (#batch, time1, size) -> (#batch, head, time1, size/head)
        k,v: (#batch, time2, size) -> (#batch, head, time2, size/head)
        p: (#batch, time2, size)
        """
        # Compute Overflows
        overflow_Q = Q.size(2) % group_size
        overflow_KV = K.size(2) % group_size

        # if-else for ONNX export
        #   0 // 0.00000000000000001 = 0
        #   1 // 1.00000000000000001 = 1
        padding_Q = (group_size - overflow_Q) * int(
            overflow_Q // (overflow_Q + 0.00000000000000001))
        padding_KV = (group_size - overflow_KV) * int(
            overflow_KV // (overflow_KV + 0.00000000000000001))

        batch_size, _, seq_len_KV, _ = K.size()

        # Input Padding (B, T, D) -> (B, T + P, D)
        Q = F.pad(Q, (0, 0, 0, padding_Q), value=0.0)
        K = F.pad(K, (0, 0, 0, padding_KV), value=0.0)
        V = F.pad(V, (0, 0, 0, padding_KV), value=0.0)

        if mask is not None and mask.size(2) > 0:  # time2 > 0:
            mask = mask[:, ::group_size, ::group_size]

        Q = Q.transpose(1, 2).contiguous().view(
            batch_size, -1, self.h, self.d_k * group_size).transpose(1, 2)
        K = K.transpose(1, 2).contiguous().view(
            batch_size, -1, self.h, self.d_k * group_size).transpose(1, 2)
        V = V.transpose(1, 2).contiguous().view(
            batch_size, -1, self.h, self.d_k * group_size).transpose(1, 2)

        # process pos_emb
        P_batch_size = P.size(0)
        overflow_P = P.size(1) % group_size
        padding_P = group_size - overflow_P if overflow_P else 0
        P = F.pad(P, (0, 0, 0, padding_P), value=0.0)
        P = P.view(P_batch_size, -1, self.h,
                   self.d_k * group_size).transpose(1, 2)

        return Q, K, V, P, mask, padding_Q

    def forward_attention(self,
                          value: torch.Tensor,
                          scores: torch.Tensor,
                          mask: torch.Tensor = torch.ones((0, 0, 0),
                                                          dtype=torch.bool),
                          padding_q: Optional[int] = None) -> torch.Tensor:
        """Compute attention context vector.

        Args:
            value (torch.Tensor): Transformed value, size
                (#batch, n_head, time2, d_k).
            scores (torch.Tensor): Attention score, size
                (#batch, n_head, time1, time2).
            mask (torch.Tensor): Mask, size (#batch, 1, time2) or
                (#batch, time1, time2), (0, 0, 0) means fake mask.
            padding_q : for GroupedAttention in efficent conformer

        Returns:
            torch.Tensor: Transformed value (#batch, time1, d_model)
                weighted by the attention score (#batch, time1, time2).

        """
        n_batch = value.size(0)
        # NOTE(xcsong): When will `if mask.size(2) > 0` be True?
        #   1. onnx(16/4) [WHY? Because we feed real cache & real mask for the
        #           1st chunk to ease the onnx export.]
        #   2. pytorch training
        if mask.size(2) > 0:  # time2 > 0
            mask = mask.unsqueeze(1).eq(0)  # (batch, 1, *, time2)
            # For last chunk, time2 might be larger than scores.size(-1)
            mask = mask[:, :, :, :scores.size(-1)]  # (batch, 1, *, time2)
            scores = scores.masked_fill(mask, -float('inf'))
            attn = torch.softmax(scores, dim=-1).masked_fill(
                mask, 0.0)  # (batch, head, time1, time2)
        # NOTE(xcsong): When will `if mask.size(2) > 0` be False?
        #   1. onnx(16/-1, -1/-1, 16/0)
        #   2. jit (16/-1, -1/-1, 16/0, 16/4)
        else:
            attn = torch.softmax(scores, dim=-1)  # (batch, head, time1, time2)

        p_attn = self.dropout(attn)
        x = torch.matmul(p_attn, value)  # (batch, head, time1, d_k)

        # n_feat!=h*d_k may be happened in GroupAttention
        x = (x.transpose(1, 2).contiguous().view(n_batch, -1, self.n_feat)
             )  # (batch, time1, d_model)
        if padding_q is not None:
            # for GroupedAttention in efficent conformer
            x = x[:, :x.size(1) - padding_q]

        return self.linear_out(x)  # (batch, time1, d_model)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.Tensor = torch.ones((0, 0, 0), dtype=torch.bool),
        pos_emb: torch.Tensor = torch.empty(0),
        cache: torch.Tensor = torch.zeros((0, 0, 0, 0)),
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute 'Scaled Dot Product Attention' with rel. positional encoding.
        Args:
            query (torch.Tensor): Query tensor (#batch, time1, size).
            key (torch.Tensor): Key tensor (#batch, time2, size).
            value (torch.Tensor): Value tensor (#batch, time2, size).
            mask (torch.Tensor): Mask tensor (#batch, 1, time2) or
                (#batch, time1, time2).
            pos_emb (torch.Tensor): Positional embedding tensor
                (#batch, time2, size).
            cache (torch.Tensor): Cache tensor (1, head, cache_t, d_k * 2),
                where `cache_t == chunk_size * num_decoding_left_chunks`
                and `head * d_k == size`
        Returns:
            torch.Tensor: Output tensor (#batch, time1, d_model).
            torch.Tensor: Cache tensor (1, head, cache_t + time1, d_k * 2)
                where `cache_t == chunk_size * num_decoding_left_chunks`
                and `head * d_k == size`
        """
        q = self.linear_q(query)
        k = self.linear_k(key)  # (#batch, time2, size)
        v = self.linear_v(value)
        p = self.linear_pos(pos_emb)  # (#batch, time2, size)

        batch_size, seq_len_KV, _ = k.size()  # seq_len_KV = time2

        # (#batch, time2, size) -> (#batch, head, time2, size/head)
        q = q.view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
        k = k.view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
        v = v.view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
        if cache.size(0) > 0:
            # use attention cache
            key_cache, value_cache = torch.split(cache,
                                                 cache.size(-1) // 2,
                                                 dim=-1)
            k = torch.cat([key_cache, k], dim=2)
            v = torch.cat([value_cache, v], dim=2)
        new_cache = torch.cat((k, v), dim=-1)

        # May be k and p does not match.  eg. time2=18+18/2=27 > mask=36/2=18
        if mask is not None and mask.size(2) > 0:
            time2 = mask.size(2)
            k = k[:, :, -time2:, :]
            v = v[:, :, -time2:, :]

        # q k v p: (batch, head, time1, d_k)
        q, k, v, p, mask, padding_q = self.pad4group(q, k, v, p, mask,
                                                     self.group_size)

        # q_with_bias_u & q_with_bias_v = (batch, head, time1, d_k)
        q = q.transpose(1, 2)  # (batch, time1, head, d_k)
        q_with_bias_u = (q + self.pos_bias_u).transpose(1, 2)
        q_with_bias_v = (q + self.pos_bias_v).transpose(1, 2)

        # compute attention score
        # first compute matrix a and matrix c
        # as described in https://arxiv.org/abs/1901.02860 Section 3.3
        # (batch, head, time1, time2)
        matrix_ac = torch.matmul(q_with_bias_u, k.transpose(-2, -1))

        # compute matrix b and matrix d
        # (batch, head, time1, time2)
        matrix_bd = torch.matmul(q_with_bias_v, p.transpose(-2, -1))
        # Remove rel_shift since it is useless in speech recognition,
        # and it requires special attention for streaming.
        # matrix_bd = self.rel_shift(matrix_bd)

        scores = (matrix_ac + matrix_bd) / math.sqrt(
            self.d_k * self.group_size)  # (batch, head, time1, time2)

        return self.forward_attention(v, scores, mask, padding_q), new_cache



# Copyright    2021  Xiaomi Corp.        (authors: Fangjun Kuang)
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
import torch.nn.functional as F
from scaling import Balancer


class Decoder(nn.Module):
    """This class modifies the stateless decoder from the following paper:

        RNN-transducer with stateless prediction network
        https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=9054419

    It removes the recurrent connection from the decoder, i.e., the prediction
    network. Different from the above paper, it adds an extra Conv1d
    right after the embedding layer.

    TODO: Implement https://arxiv.org/pdf/2109.07513.pdf
    """

    def __init__(
        self,
        vocab_size: int,
        decoder_dim: int,
        blank_id: int,
        context_size: int,
    ):
        """
        Args:
          vocab_size:
            Number of tokens of the modeling unit including blank.
          decoder_dim:
            Dimension of the input embedding, and of the decoder output.
          blank_id:
            The ID of the blank symbol.
          context_size:
            Number of previous words to use to predict the next word.
            1 means bigram; 2 means trigram. n means (n+1)-gram.
        """
        super().__init__()

        self.embedding = nn.Embedding(
            num_embeddings=vocab_size,
            embedding_dim=decoder_dim,
        )
        # the balancers are to avoid any drift in the magnitude of the
        # embeddings, which would interact badly with parameter averaging.
        self.balancer = Balancer(
            decoder_dim,
            channel_dim=-1,
            min_positive=0.0,
            max_positive=1.0,
            min_abs=0.5,
            max_abs=1.0,
            prob=0.05,
        )

        self.blank_id = blank_id

        assert context_size >= 1, context_size
        self.context_size = context_size
        self.vocab_size = vocab_size

        if context_size > 1:
            self.conv = nn.Conv1d(
                in_channels=decoder_dim,
                out_channels=decoder_dim,
                kernel_size=context_size,
                padding=0,
                groups=decoder_dim // 4,  # group size == 4
                bias=False,
            )
            self.balancer2 = Balancer(
                decoder_dim,
                channel_dim=-1,
                min_positive=0.0,
                max_positive=1.0,
                min_abs=0.5,
                max_abs=1.0,
                prob=0.05,
            )
        else:
            # To avoid `RuntimeError: Module 'Decoder' has no attribute 'conv'`
            # when inference with torch.jit.script and context_size == 1
            self.conv = nn.Identity()
            self.balancer2 = nn.Identity()

    def forward(self, y: torch.Tensor, need_pad: bool = True) -> torch.Tensor:
        """
        Args:
          y:
            A 2-D tensor of shape (N, U).
          need_pad:
            True to left pad the input. Should be True during training.
            False to not pad the input. Should be False during inference.
        Returns:
          Return a tensor of shape (N, U, decoder_dim).
        """
        y = y.to(torch.int64)
        # this stuff about clamp() is a temporary fix for a mismatch
        # at utterance start, we use negative ids in beam_search.py
        embedding_out = self.embedding(y.clamp(min=0)) * (y >= 0).unsqueeze(-1)

        embedding_out = self.balancer(embedding_out)

        if self.context_size > 1:
            embedding_out = embedding_out.permute(0, 2, 1)
            if need_pad is True:
                embedding_out = F.pad(embedding_out, pad=(self.context_size - 1, 0))
            else:
                # During inference time, there is no need to do extra padding
                # as we only need one output
                assert embedding_out.size(-1) == self.context_size
            embedding_out = self.conv(embedding_out)
            embedding_out = embedding_out.permute(0, 2, 1)
            embedding_out = F.relu(embedding_out)
            embedding_out = self.balancer2(embedding_out)

        return embedding_out



import os
import sys

import tqdm

sys.path.insert(0,'../../../')
from gxl_ai_utils.utils import utils_file
def do_remove():
    """"""
    data_dir = "/home/node36_data/xlgeng/asr_data_from_pachong/gxl_output/"
    dataset_names_list = os.listdir(data_dir)
    for dataset_name in dataset_names_list:
        utils_file.logging_print('开始处理的dataset_name:{}'.format(dataset_name))
        temp_data_dir = os.path.join(data_dir,dataset_name)
        utils_file.logging_print('先判断是否处理完毕')
        text_path = os.path.join(temp_data_dir,'text')
        text_path_2 = os.path.join(temp_data_dir,'all_2.text')
        wav_path = os.path.join(temp_data_dir,'wav.scp')
        wav_path_2 = os.path.join(temp_data_dir,'all_wav.scp')
        if (os.path.exists(text_path) and os.path.exists(wav_path)) or (os.path.exists(text_path_2) and os.path.exists(wav_path_2)):
            utils_file.logging_print('改数据集已经处理完毕,可以开启删除进程')
            now_wav_path = ""
            now_text_path = ""
            if (os.path.exists(text_path) and os.path.exists(wav_path)):
                now_text_path = text_path
                now_wav_path = wav_path
            elif (os.path.exists(text_path_2) and os.path.exists(wav_path_2)):
                now_text_path = text_path_2
                now_wav_path = wav_path_2
            temp_long_old_clean_wav_dir = os.path.join(temp_data_dir,'temp_long_old_clean_wav')
            final_wav_dir = os.path.join(temp_data_dir,'final_wav')
            utils_file.logging_print('首先 删除temp_long_old_clean_wav_dir')
            if os.path.exists(temp_long_old_clean_wav_dir):
                utils_file.remove_dir(temp_long_old_clean_wav_dir)
            utils_file.logging_print('再 删除final_wav_dir中多余的音频')
            all_wav_scp_dict = utils_file.get_scp_for_wav_dir(final_wav_dir)
            my_wav_scp_dict = utils_file.load_dict_from_scp(now_wav_path)
            utils_file.logging_print("一共存在音频数:{}".format(len(all_wav_scp_dict)))
            utils_file.logging_print("真正有效的音频数:{}".format(len(my_wav_scp_dict)))
            del_wav_dict = {}
            for k,v in all_wav_scp_dict.items():
                if k not in my_wav_scp_dict:
                    del_wav_dict[k] = v
            utils_file.logging_print('需要删除音频数:{}'.format(len(del_wav_dict)))
            if len(del_wav_dict) > 0:
                for wav_path in tqdm.tqdm(del_wav_dict.values(), desc='删除音频进度', total=len(del_wav_dict)):
                    utils_file.remove_file(wav_path)
            utils_file.logging_print('删除完毕')
        else:
            utils_file.logging_print('改数据集还未处理完毕')
            continue

if __name__ == '__main__':
    do_remove()


from gxl_ai_utils.gxl_infer_wenet.gxl_infer import GxlInterfacer

gxl_interfacer = GxlInterfacer('infer_config.yaml')
from gxl_ai_utils.utils import utils_file, utils_data

if __name__ == '__main__':
    gxl_interfacer.run_infer()



import os

import tqdm

from gxl_ai_utils.utils import utils_file


def split_wav_little_func(big_wav_dict, row_wav_dict, language, split_wav_dir, split_scp_dir):
    print('进入split_wav_little_func')
    res_wav_dict = {}
    res_text_dict = {}
    res_duration_dict = {}
    total_wav_num = len(row_wav_dict)
    i = 0
    for aid, source_wav_path in row_wav_dict.items():
        """"""
        try:
            input_audio_dict = big_wav_dict[aid]
            utils_file.logging_print(f'{language} wav schedule {i}/{total_wav_num}')
            if not os.path.exists(source_wav_path):
                utils_file.logging_print(f'{language} wav {i}/{total_wav_num} 不存在,没下载, 跳过')
                continue
            finish_file_path = source_wav_path.replace(".mp3", ".split.finish")
            if os.path.exists(finish_file_path):
                utils_file.logging_print(f'{language} wav {i}/{total_wav_num} 已经完成, 跳过')
                continue
            utils_file.logging_print('开始处理：', source_wav_path)
            normal_wav_path = source_wav_path.replace(".mp3", "_normal.wav")
            if not os.path.exists(normal_wav_path):
                utils_file.do_normalization(source_wav_path, normal_wav_path)
            else:
                utils_file.logging_print(f'normal file存在,跳过normal步骤')
            segment_list = input_audio_dict['segments']
            total_seg = len(segment_list)
            for j, segment_info_dict in enumerate(segment_list):
                utils_file.logging_print(f'wav schedule {i}/{total_wav_num}, segment schedule: {j + 1}/{total_seg}')
                sid = segment_info_dict["sid"]
                sid_name = sid + ".wav"
                begin_time = segment_info_dict["begin_time"]
                begin_time = float(begin_time)
                end_time = segment_info_dict["end_time"]
                end_time = float(end_time)
                text = segment_info_dict["text_tn"]
                utils_file.logging_print(f'{sid} {begin_time} {end_time} {text}')
                seg_wav_path = os.path.join(split_wav_dir, sid_name)
                sample_rate = 16000
                utils_file.do_extract_audio_segment(normal_wav_path, seg_wav_path, int(begin_time * sample_rate),
                                                    int(end_time * sample_rate))
                res_wav_dict[sid] = seg_wav_path
                res_text_dict[sid] = text
                res_duration_dict[sid] = end_time - begin_time
            utils_file.remove_file(normal_wav_path)
            utils_file.write_list_to_file([],finish_file_path)
        except Exception as e:
            utils_file.logging_print(f'{e}')
        i += 1
    utils_file.write_dict_to_scp(res_wav_dict, os.path.join(split_scp_dir, 'wav.scp'))
    utils_file.write_dict_to_scp(res_text_dict, os.path.join(split_scp_dir, 'text'))
    utils_file.write_dict_to_scp(res_duration_dict, os.path.join(split_scp_dir, 'duration.scp'))
def split_wav(language, wav_dict_list):
    """"""
    split_wav_dir = f"/home/work_nfs13/xlgeng/data/MSR-86K/MSR-86K/split_wav_new/{language}"
    split_scp_dir = f"/home/work_nfs13/xlgeng/data/MSR-86K/MSR-86K/split_scp_new/{language}"
    utils_file.makedir(split_wav_dir)
    utils_file.makedir(split_scp_dir)

    original_wav_dict_path = os.path.join(split_scp_dir, 'origin_wav.scp')
    row_wav_dir = "/home/work_nfs13/xlgeng/data/MSR-86K/MSR-86K/row_wav"
    row_dir_language = os.path.join(row_wav_dir, language)
    row_wav_dict = utils_file.do_get_scp_for_wav_dir(row_dir_language, suffix='.mp3')
    utils_file.write_dict_to_scp(row_wav_dict, original_wav_dict_path)
    utils_file.logging_print(f"{language}:  now original_wav num: {len(row_wav_dict)}, should original_wav num: {len(wav_dict_list)}")
    big_wav_dict = {}
    for dict_item in tqdm.tqdm(wav_dict_list, desc=f"{language}_big_dict", total=len(wav_dict_list)):
        big_wav_dict[dict_item['aid']] = dict_item

    num_thread = 32
    runner = utils_file.GxlDynamicProcessPool()
    row_wav_dict_list = utils_file.do_split_dict(row_wav_dict, num_thread)
    index = 0
    for row_wav_dict_i in row_wav_dict_list:
        split_scp_dir_i = os.path.join(split_scp_dir, f'split_{index}')
        split_wav_dir_i = os.path.join(split_wav_dir, f'split_{index}')
        utils_file.makedir(split_scp_dir_i)
        utils_file.makedir(split_wav_dir_i)
        runner.add_task(split_wav_little_func, [big_wav_dict, row_wav_dict_i, language, split_wav_dir_i, split_scp_dir_i])
        index += 1
    runner.run()


def main():
    """"""
    root_dir = "/home/work_nfs13/xlgeng/data/MSR-86K/MSR-86K/row_wav"
    utils_file.makedir(root_dir)
    language = "Thai"
    utils_file.logging_print(language)
    input_json_dir = "/home/work_nfs13/xlgeng/data/MSR-86K/MSR-86K"
    input_json_dir_language = os.path.join(input_json_dir, language)
    dev_json_path = os.path.join(input_json_dir_language, 'dev.json')
    train_json_path = os.path.join(input_json_dir_language, 'train.json')
    if not os.path.exists(dev_json_path) or not os.path.exists(train_json_path):
        utils_file.logging_print(f'{language} 不存在dev.json或train.json')
        return
    dev_json_dict = utils_file.load_dict_from_json(dev_json_path)
    train_json_dict = utils_file.load_dict_from_json(train_json_path)
    wav_dict_list = dev_json_dict['audios'] + train_json_dict['audios']
    split_wav(language, wav_dict_list)

if __name__ == '__main__':
    main()


from typing import Tuple, Union
import torch
from wenet.transformer.subsampling import BaseSubsampling


class IdentitySubsampling(BaseSubsampling):
    """ Paraformer subsampling
    """

    def __init__(self, idim: int, odim: int, dropout_rate: float,
                 pos_enc_class: torch.nn.Module):
        super().__init__()
        _, _ = idim, odim
        self.right_context = 6
        self.subsampling_rate = 6
        self.pos_enc = pos_enc_class

    def forward(
        self,
        x: torch.Tensor,
        x_mask: torch.Tensor,
        offset: Union[torch.Tensor, int] = 0
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Subsample x.

        Args:
            x (torch.Tensor): Input tensor (#batch, time, idim).
            x_mask (torch.Tensor): Input mask (#batch, 1, time).

        Returns:
            torch.Tensor: Subsampled tensor (#batch, time', odim),
                where time' = time.
            torch.Tensor: Subsampled mask (#batch, 1, time'),
                where time' = time
            torch.Tensor: positional encoding

        """
        # NOTE(Mddct): Paraformer starts from 1
        if isinstance(offset, torch.Tensor):
            offset = torch.add(offset, 1)
        else:
            offset = offset + 1
        x, pos_emb = self.pos_enc(x, offset)
        return x, pos_emb, x_mask

    def position_encoding(self, offset: Union[int, torch.Tensor],
                          size: int) -> torch.Tensor:
        return self.pos_enc.position_encoding(offset + 1, size)



import logging
import sys
from pathlib import Path
from typing import Dict, List

import k2
import torch
import tqdm
import os
# 获取当前工作目录
current_directory = "/home/work_nfs8/xlgeng/new_workspace/wenet_gxl_aishell/examples/aishell/do_goutu"
# 添加指定目录到PYTHONPATH
python_path = os.path.join(current_directory, "tools", "k2", "icefall")
python_path_2 = os.path.join(current_directory, "tools", "k2", "icefall", "egs", "aishell", "ASR", "local")
# 检查PYTHONPATH是否已设置，如果没有则设置为空字符串
if "PYTHONPATH" not in os.environ:
    os.environ["PYTHONPATH"] = ""
# 将指定目录添加到PYTHONPATH
os.environ["PYTHONPATH"] += ":" + python_path
os.environ["PYTHONPATH"] += ":" + python_path_2

from prepare_lang import (
    Lexicon,
    add_disambig_symbols,
    add_self_loops,
    write_lexicon,
    write_mapping,
)
from gxl_ai_utils.utils import utils_file


def lexicon_to_fst_no_sil(
    lexicon: Lexicon,
    token2id: Dict[str, int],
    word2id: Dict[str, int],
    need_self_loops: bool = False,
) -> k2.Fsa:
    """Convert a lexicon to an FST (in k2 format).

    Args:
      lexicon:
        The input lexicon. See also :func:`read_lexicon`
      token2id:
        A dict mapping tokens to IDs.
      word2id:
        A dict mapping words to IDs.
      need_self_loops:
        If True, add self-loop to states with non-epsilon output symbols
        on at least one arc out of the state. The input label for this
        self loop is `token2id["#0"]` and the output label is `word2id["#0"]`.
    Returns:
      Return an instance of `k2.Fsa` representing the given lexicon.
    """
    loop_state = 0  # words enter and leave from here
    next_state = 1  # the next un-allocated state, will be incremented as we go

    arcs = []

    # The blank symbol <blk> is defined in local/train_bpe_model.py
    assert token2id["<blank>"] == 0
    assert word2id["<eps>"] == 0

    eps = 0

    for word, pieces in lexicon:
        assert len(pieces) > 0, f"{word} has no pronunciations"
        cur_state = loop_state

        word = word2id[word]
        pieces = [
            token2id[i] if i in token2id else token2id["<unk>"] for i in pieces
        ]

        for i in range(len(pieces) - 1):
            w = word if i == 0 else eps
            arcs.append([cur_state, next_state, pieces[i], w, 0])

            cur_state = next_state
            next_state += 1

        # now for the last piece of this word
        i = len(pieces) - 1
        w = word if i == 0 else eps
        arcs.append([cur_state, loop_state, pieces[i], w, 0])

    if need_self_loops:
        disambig_token = token2id["#0"]
        disambig_word = word2id["#0"]
        arcs = add_self_loops(
            arcs,
            disambig_token=disambig_token,
            disambig_word=disambig_word,
        )

    final_state = next_state
    arcs.append([loop_state, final_state, -1, -1, 0])
    arcs.append([final_state])

    arcs = sorted(arcs, key=lambda arc: arc[0])
    arcs = [[str(i) for i in arc] for arc in arcs]
    arcs = [" ".join(arc) for arc in arcs]
    arcs = "\n".join(arcs)

    fsa = k2.Fsa.from_str(arcs, acceptor=False)
    return fsa


def contain_oov(token_sym_table: Dict[str, int], tokens: List[str]) -> bool:
    """Check if all the given tokens are in token symbol table.

    Args:
      token_sym_table:
        Token symbol table that contains all the valid tokens.
      tokens:
        A list of tokens.
    Returns:
      Return True if there is any token not in the token_sym_table,
      otherwise False.
    """
    for tok in tokens:
        if tok not in token_sym_table:
            return True
    return False


def generate_lexicon(token_sym_table: Dict[str, int],
                     words: List[str]) -> Lexicon:
    """Generate a lexicon from a word list and token_sym_table.

    Args:
      token_sym_table:
        Token symbol table that mapping token to token ids.
      words:
        A list of strings representing words.
    Returns:
      Return a dict whose keys are words and values are the corresponding
          tokens.
    """
    lexicon = []
    for word in tqdm.tqdm(words, desc="Generating lexicon", total=len(words)):
        chars = list(word.strip(" \t"))
        if contain_oov(token_sym_table, chars):
            continue
        lexicon.append((word, chars))

    # The OOV word is <UNK>
    lexicon.append(("<UNK>", ["<unk>"]))
    return lexicon


def generate_tokens(text_file: str) -> Dict[str, int]:
    """Generate tokens from the given text file.

    Args:
      text_file:
        A file that contains text lines to generate tokens.
    Returns:
      Return a dict whose keys are tokens and values are token ids ranged
      from 0 to len(keys) - 1.
    """
    token2id: Dict[str, int] = dict()
    with open(text_file, "r", encoding="utf-8") as f:
        for line in f:
            char, index = line.replace('\n', '').split()
            assert char not in token2id
            token2id[char] = int(index)
    assert token2id['<blank>'] == 0
    return token2id


def generate_words(text_file: str) -> Dict[str, int]:
    """Generate words from the given text file.

    Args:
      text_file:
        A file that contains text lines to generate words.
    Returns:
      Return a dict whose keys are words and values are words ids ranged
      from 0 to len(keys) - 1.
    """
    words = []
    lines = utils_file.load_list_file_clean(text_file)
    for line in tqdm.tqdm(lines, desc="Generating words", total=len(lines)):
        word = line.replace('\n', '').split()[0]
        # assert word not in words
        words.append(word)
    words = list(set(words))
    words.sort()
    utils_file.logging_print('耿雪龙: 生成词典list len: {}'.format(len(words)))

    # We put '<eps>' '<UNK>' at begining of word2id
    # '#0', '<s>', '</s>' at end of word2id
    words = [
        word for word in words
        if word not in ['<eps>', '<UNK>', '#0', '<s>', '</s>']
    ]
    words.insert(0, '<eps>')
    words.insert(1, '<UNK>')
    words.append('#0')
    words.append('<s>')
    words.append('</s>')
    word2id = {j: i for i, j in enumerate(words)}
    return word2id

def prepare_char(input_words_path, input_tokens_path, output_dir):
    utils_file.makedir(output_dir)
    timer = utils_file.GxlTimer()
    timer.start()
    utils_file.logging_print('耿雪龙: 开始生成字典')
    token2id = generate_tokens(input_tokens_path)
    utils_file.logging_print('耿雪龙: 生成字典完成, len: {}'.format(len(token2id)))
    timer.stop_halfway()
    utils_file.logging_print('耿雪龙: 开始生成word2id')
    word2id = generate_words(input_words_path)
    utils_file.logging_print('耿雪龙: 生成word2id完成, len: {}'.format(len(word2id)))
    timer.stop_halfway()
    tgt_dir = Path(output_dir)

    words = [
        word for word in word2id.keys() if word not in
        ["<eps>", "!SIL", "<SPOKEN_NOISE>", "<UNK>", "#0", "<s>", "</s>"]
    ]
    utils_file.logging_print('耿雪龙: 开始生成词典')
    lexicon = generate_lexicon(token2id, words)
    utils_file.logging_print('Lexicon size: {}'.format(len(lexicon)))
    timer.stop_halfway()

    lexicon_disambig, max_disambig = add_disambig_symbols(lexicon)
    next_token_id = max(token2id.values()) + 1
    for i in range(max_disambig + 1):
        disambig = f"#{i}"
        assert disambig not in token2id
        token2id[disambig] = next_token_id
        next_token_id += 1

    write_mapping(tgt_dir / "tokens.txt", token2id)
    write_mapping(tgt_dir / "words.txt", word2id)
    write_lexicon(tgt_dir / "lexicon.txt", lexicon)
    write_lexicon(tgt_dir / "lexicon_disambig.txt", lexicon_disambig)

    L = lexicon_to_fst_no_sil(
        lexicon,
        token2id=token2id,
        word2id=word2id,
    )
    L_disambig = lexicon_to_fst_no_sil(
        lexicon_disambig,
        token2id=token2id,
        word2id=word2id,
        need_self_loops=True,
    )
    torch.save(L.as_dict(), tgt_dir / "L.pt")
    torch.save(L_disambig.as_dict(), tgt_dir / "L_disambig.pt")

if __name__ == "__main__":
    token_file = "/home/work_nfs7/yhliang/online_sys/wenet/examples/onlinesys/data/dict/lang_char.txt"
    word_file = "/home/work_nfs4_ssd/yhliang/LanguageModel/lang/lang_test616_mix0.1_pruned_5e-11/words.txt"
    output_dir = "./output_data/"
    utils_file.logging_print('哈哈哈哈')
    prepare_char(input_tokens_path=token_file, input_words_path=word_file, output_dir=output_dir)



import tqdm
import sys
sys.path.insert(0, '/home/work_nfs11/code/xlgeng/gxl_ai_utils')

from gxl_ai_utils.utils import utils_file
utils_file.do_convert_shards2raw()
from tmp_utils import do_get_formatted_datalist_for_asr_task
# 把测试集 从 nfs15搬运到 54本地盘

input_dir = "/home/work_nfs15/asr_data/data/asr_test_sets"
output_dir = "/home/node54_tmpdata/xlgeng/asr_test_sets"
dataset_names,_ = utils_file.do_listdir(input_dir, return_path=False)

def little_func_for_cp(input_wav_dict, output_dir, index ):
    if index == 0:
        for key, wav_path in tqdm.tqdm(input_wav_dict.items(), total=len(input_wav_dict), desc='复制文件'):
            utils_file.copy_file2(wav_path, output_dir, is_jump=True, is_log=False)
    else:
        for key, wav_path in input_wav_dict.items():
            utils_file.copy_file2(wav_path, output_dir, is_jump=True, is_log=False)


for dataset_name in dataset_names:
    utils_file.logging_print('开始处理', dataset_name)
    multi_process_runner = utils_file.GxlFixedProcessPool(20)
    input_dir_tmp = utils_file.join_path(input_dir, dataset_name)
    wav_dir = utils_file.join_path(input_dir_tmp, "wav")
    wav_scp_path = utils_file.join_path(input_dir_tmp, "wav.scp")
    text_path = utils_file.join_path(input_dir_tmp, "text")
    if not utils_file.if_file_exist(wav_scp_path)  or not utils_file.if_dir_exist(wav_dir) or not utils_file.if_file_exist(text_path):
        utils_file.logging_print("skip:", dataset_name, '因为wav.scp text or wav dir不存在')
        continue
    output_wav_dir = utils_file.join_path(output_dir, dataset_name, 'wav')
    utils_file.makedir_sil(output_wav_dir)
    output_wav_scp_path = utils_file.join_path(output_dir, dataset_name, 'wav.scp')
    output_text_path = utils_file.join_path(output_dir, dataset_name, 'text')
    output_data_list_path = utils_file.join_path(output_dir, dataset_name, 'data.list')
    wav_path_dict = utils_file.load_dict_from_scp(wav_scp_path)
    # 进行文件复制
    utils_file.logging_print('开始进行文件复制, 使用20个进程')
    numb_dict_list = utils_file.do_split_dict(wav_path_dict, 20)
    for i, dict_i in enumerate(numb_dict_list):
        multi_process_runner.add_thread(little_func_for_cp, [dict_i, output_wav_dir, i])
    multi_process_runner.start()
    utils_file.logging_print('复制完成')
    utils_file.logging_print('开始得到data.list')
    new_wav_path_dict = {}
    for key, wav_path in wav_path_dict.items():
        new_wav_path = utils_file.do_replace_dir(wav_path, output_wav_dir)
        new_wav_path_dict[key] = new_wav_path
    text_dict = utils_file.load_dict_from_scp(text_path)
    res_dict_list = do_get_formatted_datalist_for_asr_task(new_wav_path_dict, text_dict, dataset_name)
    utils_file.write_dict_list_to_jsonl(res_dict_list, output_data_list_path)
    utils_file.write_dict_to_scp(new_wav_path_dict, output_wav_scp_path)
    utils_file.write_dict_to_scp(text_dict, output_text_path)





#!/usr/bin/env python3
#
# Copyright 2021-2023 Xiaomi Corporation (Author: Fangjun Kuang,
#                                                 Zengwei Yao,
#                                                 Wei Kang)
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This script converts several saved checkpoints
# to a single one using model averaging.
"""

Usage:

Note: This is a example for librispeech dataset, if you are using different
dataset, you should change the argument values according to your dataset.

(1) Export to torchscript model using torch.jit.script()

- For non-streaming model:

./zipformer/export.py \
  --exp-dir ./zipformer/exp \
  --tokens data/lang_bpe_500/tokens.txt \
  --epoch 30 \
  --avg 9 \
  --jit 1

It will generate a file `jit_script.pt` in the given `exp_dir`. You can later
load it by `torch.jit.load("jit_script.pt")`.

Check ./jit_pretrained.py for its usage.

Check https://github.com/k2-fsa/sherpa
for how to use the exported models outside of icefall.

- For streaming model:

./zipformer/export.py \
  --exp-dir ./zipformer/exp \
  --causal 1 \
  --chunk-size 16 \
  --left-context-frames 128 \
  --tokens data/lang_bpe_500/tokens.txt \
  --epoch 30 \
  --avg 9 \
  --jit 1

It will generate a file `jit_script_chunk_16_left_128.pt` in the given `exp_dir`.
You can later load it by `torch.jit.load("jit_script_chunk_16_left_128.pt")`.

Check ./jit_pretrained_streaming.py for its usage.

Check https://github.com/k2-fsa/sherpa
for how to use the exported models outside of icefall.

(2) Export `model.state_dict()`

- For non-streaming model:

./zipformer/export.py \
  --exp-dir ./zipformer/exp \
  --tokens data/lang_bpe_500/tokens.txt \
  --epoch 30 \
  --avg 9

- For streaming model:

./zipformer/export.py \
  --exp-dir ./zipformer/exp \
  --causal 1 \
  --tokens data/lang_bpe_500/tokens.txt \
  --epoch 30 \
  --avg 9

It will generate a file `pretrained.pt` in the given `exp_dir`. You can later
load it by `icefall.checkpoint.load_checkpoint()`.

- For non-streaming model:

To use the generated file with `zipformer/decode.py`,
you can do:

    cd /path/to/exp_dir
    ln -s pretrained.pt epoch-9999.pt

    cd /path/to/egs/librispeech/ASR
    ./zipformer/decode.py \
        --exp-dir ./zipformer/exp \
        --epoch 9999 \
        --avg 1 \
        --max-duration 600 \
        --decoding-method greedy_search \
        --bpe-model data/lang_bpe_500/bpe.model

- For streaming model:

To use the generated file with `zipformer/decode.py` and `zipformer/streaming_decode.py`, you can do:

    cd /path/to/exp_dir
    ln -s pretrained.pt epoch-9999.pt

    cd /path/to/egs/librispeech/ASR

    # simulated streaming decoding
    ./zipformer/decode.py \
        --exp-dir ./zipformer/exp \
        --epoch 9999 \
        --avg 1 \
        --max-duration 600 \
        --causal 1 \
        --chunk-size 16 \
        --left-context-frames 128 \
        --decoding-method greedy_search \
        --bpe-model data/lang_bpe_500/bpe.model

    # chunk-wise streaming decoding
    ./zipformer/streaming_decode.py \
        --exp-dir ./zipformer/exp \
        --epoch 9999 \
        --avg 1 \
        --max-duration 600 \
        --causal 1 \
        --chunk-size 16 \
        --left-context-frames 128 \
        --decoding-method greedy_search \
        --bpe-model data/lang_bpe_500/bpe.model

Check ./pretrained.py for its usage.

Note: If you don't want to train a model from scratch, we have
provided one for you. You can get it at

- non-streaming model:
https://huggingface.co/Zengwei/icefall-asr-librispeech-zipformer-2023-05-15

- streaming model:
https://huggingface.co/Zengwei/icefall-asr-librispeech-streaming-zipformer-2023-05-17

with the following commands:

    sudo apt-get install git-lfs
    git lfs install
    git clone https://huggingface.co/Zengwei/icefall-asr-librispeech-zipformer-2023-05-15
    git clone https://huggingface.co/Zengwei/icefall-asr-librispeech-streaming-zipformer-2023-05-17
    # You will find the pre-trained models in exp dir
"""

import argparse
import logging
from pathlib import Path
from typing import List, Tuple

import k2
import torch
from scaling_converter import convert_scaled_to_non_scaled
from torch import Tensor, nn
from train import add_model_arguments, get_model, get_params

from icefall.checkpoint import (
    average_checkpoints,
    average_checkpoints_with_averaged_model,
    find_checkpoints,
    load_checkpoint,
)
from icefall.utils import make_pad_mask, num_tokens, str2bool


def get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "--epoch",
        type=int,
        default=30,
        help="""It specifies the checkpoint to use for decoding.
        Note: Epoch counts from 1.
        You can specify --avg to use more checkpoints for model averaging.""",
    )

    parser.add_argument(
        "--iter",
        type=int,
        default=0,
        help="""If positive, --epoch is ignored and it
        will use the checkpoint exp_dir/checkpoint-iter.pt.
        You can specify --avg to use more checkpoints for model averaging.
        """,
    )

    parser.add_argument(
        "--avg",
        type=int,
        default=9,
        help="Number of checkpoints to average. Automatically select "
        "consecutive checkpoints before the checkpoint specified by "
        "'--epoch' and '--iter'",
    )

    parser.add_argument(
        "--use-averaged-model",
        type=str2bool,
        default=True,
        help="Whether to load averaged model. Currently it only supports "
        "using --epoch. If True, it would decode with the averaged model "
        "over the epoch range from `epoch-avg` (excluded) to `epoch`."
        "Actually only the models with epoch number of `epoch-avg` and "
        "`epoch` are loaded for averaging. ",
    )

    parser.add_argument(
        "--exp-dir",
        type=str,
        default="zipformer/exp",
        help="""It specifies the directory where all training related
        files, e.g., checkpoints, log, etc, are saved
        """,
    )

    parser.add_argument(
        "--tokens",
        type=str,
        default="data/lang_bpe_500/tokens.txt",
        help="Path to the tokens.txt",
    )

    parser.add_argument(
        "--jit",
        type=str2bool,
        default=False,
        help="""True to save a model after applying torch.jit.script.
        It will generate a file named jit_script.pt.
        Check ./jit_pretrained.py for how to use it.
        """,
    )

    parser.add_argument(
        "--context-size",
        type=int,
        default=2,
        help="The context size in the decoder. 1 means bigram; 2 means tri-gram",
    )

    add_model_arguments(parser)

    return parser


class EncoderModel(nn.Module):
    """A wrapper for encoder and encoder_embed"""

    def __init__(self, encoder: nn.Module, encoder_embed: nn.Module) -> None:
        super().__init__()
        self.encoder = encoder
        self.encoder_embed = encoder_embed

    def forward(
        self, features: Tensor, feature_lengths: Tensor
    ) -> Tuple[Tensor, Tensor]:
        """
        Args:
            features: (N, T, C)
            feature_lengths: (N,)
        """
        x, x_lens = self.encoder_embed(features, feature_lengths)

        src_key_padding_mask = make_pad_mask(x_lens)
        x = x.permute(1, 0, 2)  # (N, T, C) -> (T, N, C)

        encoder_out, encoder_out_lens = self.encoder(x, x_lens, src_key_padding_mask)
        encoder_out = encoder_out.permute(1, 0, 2)  # (T, N, C) ->(N, T, C)

        return encoder_out, encoder_out_lens


class StreamingEncoderModel(nn.Module):
    """A wrapper for encoder and encoder_embed"""

    def __init__(self, encoder: nn.Module, encoder_embed: nn.Module) -> None:
        super().__init__()
        assert len(encoder.chunk_size) == 1, encoder.chunk_size
        assert len(encoder.left_context_frames) == 1, encoder.left_context_frames
        self.chunk_size = encoder.chunk_size[0]
        self.left_context_len = encoder.left_context_frames[0]

        # The encoder_embed subsample features (T - 7) // 2
        # The ConvNeXt module needs (7 - 1) // 2 = 3 frames of right padding after subsampling
        self.pad_length = 7 + 2 * 3

        self.encoder = encoder
        self.encoder_embed = encoder_embed

    def forward(
        self, features: Tensor, feature_lengths: Tensor, states: List[Tensor]
    ) -> Tuple[Tensor, Tensor, List[Tensor]]:
        """Streaming forward for encoder_embed and encoder.

        Args:
            features: (N, T, C)
            feature_lengths: (N,)
            states: a list of Tensors

        Returns encoder outputs, output lengths, and updated states.
        """
        chunk_size = self.chunk_size
        left_context_len = self.left_context_len

        cached_embed_left_pad = states[-2]
        x, x_lens, new_cached_embed_left_pad = self.encoder_embed.streaming_forward(
            x=features,
            x_lens=feature_lengths,
            cached_left_pad=cached_embed_left_pad,
        )
        assert x.size(1) == chunk_size, (x.size(1), chunk_size)

        src_key_padding_mask = make_pad_mask(x_lens)

        # processed_mask is used to mask out initial states
        processed_mask = torch.arange(left_context_len, device=x.device).expand(
            x.size(0), left_context_len
        )
        processed_lens = states[-1]  # (batch,)
        # (batch, left_context_size)
        processed_mask = (processed_lens.unsqueeze(1) <= processed_mask).flip(1)
        # Update processed lengths
        new_processed_lens = processed_lens + x_lens

        # (batch, left_context_size + chunk_size)
        src_key_padding_mask = torch.cat([processed_mask, src_key_padding_mask], dim=1)

        x = x.permute(1, 0, 2)  # (N, T, C) -> (T, N, C)
        encoder_states = states[:-2]

        (
            encoder_out,
            encoder_out_lens,
            new_encoder_states,
        ) = self.encoder.streaming_forward(
            x=x,
            x_lens=x_lens,
            states=encoder_states,
            src_key_padding_mask=src_key_padding_mask,
        )
        encoder_out = encoder_out.permute(1, 0, 2)  # (T, N, C) ->(N, T, C)

        new_states = new_encoder_states + [
            new_cached_embed_left_pad,
            new_processed_lens,
        ]
        return encoder_out, encoder_out_lens, new_states

    @torch.jit.export
    def get_init_states(
        self,
        batch_size: int = 1,
        device: torch.device = torch.device("cpu"),
    ) -> List[torch.Tensor]:
        """
        Returns a list of cached tensors of all encoder layers. For layer-i, states[i*6:(i+1)*6]
        is (cached_key, cached_nonlin_attn, cached_val1, cached_val2, cached_conv1, cached_conv2).
        states[-2] is the cached left padding for ConvNeXt module,
        of shape (batch_size, num_channels, left_pad, num_freqs)
        states[-1] is processed_lens of shape (batch,), which records the number
        of processed frames (at 50hz frame rate, after encoder_embed) for each sample in batch.
        """
        states = self.encoder.get_init_states(batch_size, device)

        embed_states = self.encoder_embed.get_init_states(batch_size, device)
        states.append(embed_states)

        processed_lens = torch.zeros(batch_size, dtype=torch.int32, device=device)
        states.append(processed_lens)

        return states


@torch.no_grad()
def main():
    args = get_parser().parse_args()
    args.exp_dir = Path(args.exp_dir)

    params = get_params()
    params.update(vars(args))

    device = torch.device("cpu")
    # if torch.cuda.is_available():
    #     device = torch.device("cuda", 0)

    logging.info(f"device: {device}")

    token_table = k2.SymbolTable.from_file(params.tokens)
    params.blank_id = token_table["<blk>"]
    params.vocab_size = num_tokens(token_table) + 1

    logging.info(params)

    logging.info("About to create model")
    model = get_model(params)

    if not params.use_averaged_model:
        if params.iter > 0:
            filenames = find_checkpoints(params.exp_dir, iteration=-params.iter)[
                : params.avg
            ]
            if len(filenames) == 0:
                raise ValueError(
                    f"No checkpoints found for"
                    f" --iter {params.iter}, --avg {params.avg}"
                )
            elif len(filenames) < params.avg:
                raise ValueError(
                    f"Not enough checkpoints ({len(filenames)}) found for"
                    f" --iter {params.iter}, --avg {params.avg}"
                )
            logging.info(f"averaging {filenames}")
            model.load_state_dict(average_checkpoints(filenames, device=device))
        elif params.avg == 1:
            load_checkpoint(f"{params.exp_dir}/epoch-{params.epoch}.pt", model)
        else:
            start = params.epoch - params.avg + 1
            filenames = []
            for i in range(start, params.epoch + 1):
                if i >= 1:
                    filenames.append(f"{params.exp_dir}/epoch-{i}.pt")
            logging.info(f"averaging {filenames}")
            model.load_state_dict(average_checkpoints(filenames, device=device))
    else:
        if params.iter > 0:
            filenames = find_checkpoints(params.exp_dir, iteration=-params.iter)[
                : params.avg + 1
            ]
            if len(filenames) == 0:
                raise ValueError(
                    f"No checkpoints found for"
                    f" --iter {params.iter}, --avg {params.avg}"
                )
            elif len(filenames) < params.avg + 1:
                raise ValueError(
                    f"Not enough checkpoints ({len(filenames)}) found for"
                    f" --iter {params.iter}, --avg {params.avg}"
                )
            filename_start = filenames[-1]
            filename_end = filenames[0]
            logging.info(
                "Calculating the averaged model over iteration checkpoints"
                f" from {filename_start} (excluded) to {filename_end}"
            )
            model.load_state_dict(
                average_checkpoints_with_averaged_model(
                    filename_start=filename_start,
                    filename_end=filename_end,
                    device=device,
                )
            )
        elif params.avg == 1:
            load_checkpoint(f"{params.exp_dir}/epoch-{params.epoch}.pt", model)
        else:
            assert params.avg > 0, params.avg
            start = params.epoch - params.avg
            assert start >= 1, start
            filename_start = f"{params.exp_dir}/epoch-{start}.pt"
            filename_end = f"{params.exp_dir}/epoch-{params.epoch}.pt"
            logging.info(
                f"Calculating the averaged model over epoch range from "
                f"{start} (excluded) to {params.epoch}"
            )
            model.load_state_dict(
                average_checkpoints_with_averaged_model(
                    filename_start=filename_start,
                    filename_end=filename_end,
                    device=device,
                )
            )

    model.eval()

    if params.jit is True:
        convert_scaled_to_non_scaled(model, inplace=True)
        # We won't use the forward() method of the model in C++, so just ignore
        # it here.
        # Otherwise, one of its arguments is a ragged tensor and is not
        # torch scriptabe.
        model.__class__.forward = torch.jit.ignore(model.__class__.forward)

        # Wrap encoder and encoder_embed as a module
        if params.causal:
            model.encoder = StreamingEncoderModel(model.encoder, model.encoder_embed)
            chunk_size = model.encoder.chunk_size
            left_context_len = model.encoder.left_context_len
            filename = f"jit_script_chunk_{chunk_size}_left_{left_context_len}.pt"
        else:
            model.encoder = EncoderModel(model.encoder, model.encoder_embed)
            filename = "jit_script.pt"

        logging.info("Using torch.jit.script")
        model = torch.jit.script(model)
        model.save(str(params.exp_dir / filename))
        logging.info(f"Saved to {filename}")
    else:
        logging.info("Not using torchscript. Export model.state_dict()")
        # Save it using a format so that it can be loaded
        # by :func:`load_checkpoint`
        filename = params.exp_dir / "pretrained.pt"
        torch.save({"model": model.state_dict()}, str(filename))
        logging.info(f"Saved to {filename}")


if __name__ == "__main__":
    formatter = "%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"

    logging.basicConfig(format=formatter, level=logging.INFO)
    main()



import os
import sys
sys.path.insert(0,'../../../../')
from gxl_ai_utils.utils import utils_file

wav_path = "/home/work_nfs8/xlgeng/new_workspace/data/scp_asr/aishell4/wav.scp"
text_path = "/home/work_nfs8/xlgeng/new_workspace/data/scp_asr/aishell4/text"
output_dir = "/home/work_nfs10/asr_data/data/aishell4/shards"
utils_file.makedir_sil(output_dir)
now_time = utils_file.do_get_now_time()
utils_file.do_make_shard_file(wav_path, text_path, output_dir)
print("done")
time_elapsed = utils_file.do_get_elapsed_time(now_time)
utils_file.logging_info(f'耗时：{time_elapsed} s')



# Copyright (c) 2023 Wenet Community. (authors: Dinghao Zhou)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from collections.abc import Callable
import sys
import tarfile
import logging
from typing import List
import torch
from torch.utils.data import IterDataPipe, functional_datapipe
from torch.utils.data import datapipes
from torch.utils.data.datapipes.iter import Mapper
from torch.utils.data.datapipes.iter.sharding import (
    SHARDING_PRIORITIES, ShardingFilterIterDataPipe)
from torch.utils.data.datapipes.utils.common import _check_unpickable_fn

from wenet.dataset.processor import parse_url


@functional_datapipe("map_ignore_error")
class MapperIgnoreErrorDataPipe(Mapper):

    def __init__(self,
                 dataset: IterDataPipe,
                 fn: Callable,
                 input_col=None,
                 output_col=None,
                 log_error: bool = True) -> None:
        super().__init__(dataset, fn, input_col, output_col)
        self._iter = None
        self.log_error = log_error

    def __iter__(self):
        if self._iter is None:
            self._iter = iter(self.datapipe)

        while True:
            try:
                elem = next(self._iter)
                yield self._apply_fn(elem)
            except StopIteration:
                self._iter = None
                return
            except Exception as ex:
                if self.log_error:
                    logging.warning(str(ex))


@functional_datapipe('bucket_by_sequence_length')
class BucketBySequenceLengthDataPipe(IterDataPipe):

    def __init__(
        self,
        dataset: IterDataPipe,
        elem_length_func,
        bucket_boundaries: List[int],
        bucket_batch_sizes: List[int],
        wrapper_class=None,
    ) -> None:
        super().__init__()
        _check_unpickable_fn(elem_length_func)
        assert len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        self.bucket_batch_sizes = bucket_batch_sizes
        self.bucket_boundaries = bucket_boundaries + [sys.maxsize]
        self.elem_length_func = elem_length_func

        self._group_dp = GroupByWindowDataPipe(dataset,
                                               self._element_to_bucket_id,
                                               self._window_size_func,
                                               wrapper_class=wrapper_class)

    def __iter__(self):
        yield from self._group_dp

    def _element_to_bucket_id(self, elem):
        seq_len = self.elem_length_func(elem)
        bucket_id = 0
        for (i, b) in enumerate(self.bucket_boundaries):
            if seq_len < b:
                bucket_id = i
                break
        return bucket_id

    def _window_size_func(self, bucket_id):
        return self.bucket_batch_sizes[bucket_id]


@functional_datapipe("group_by_window")
class GroupByWindowDataPipe(datapipes.iter.Grouper):

    def __init__(
        self,
        dataset: IterDataPipe,
        key_func,
        window_size_func,
        wrapper_class=None,
    ):
        super().__init__(dataset,
                         key_func,
                         keep_key=False,
                         group_size=None,
                         drop_remaining=False)
        _check_unpickable_fn(window_size_func)
        self.dp = dataset
        self.window_size_func = window_size_func
        if wrapper_class is not None:
            _check_unpickable_fn(wrapper_class)
            del self.wrapper_class
            self.wrapper_class = wrapper_class

    def __iter__(self):
        for x in self.datapipe:
            key = self.group_key_fn(x)

            self.buffer_elements[key].append(x)
            self.curr_buffer_size += 1

            group_size = self.window_size_func(key)
            if group_size == len(self.buffer_elements[key]):
                result = self.wrapper_class(self.buffer_elements[key])
                yield result
                self.curr_buffer_size -= len(self.buffer_elements[key])
                del self.buffer_elements[key]

            if self.curr_buffer_size == self.max_buffer_size:
                result_to_yield = self._remove_biggest_key()
                if result_to_yield is not None:
                    result = self.wrapper_class(result_to_yield)
                    yield result

        for key in tuple(self.buffer_elements.keys()):
            result = self.wrapper_class(self.buffer_elements.pop(key))
            self.curr_buffer_size -= len(result)
            yield result


@functional_datapipe("sort")
class SortDataPipe(IterDataPipe):

    def __init__(self,
                 dataset: IterDataPipe,
                 buffer_size: int = 500,
                 key_func=None,
                 reverse=False) -> None:
        if key_func is not None:
            _check_unpickable_fn(key_func)
        self.buffer_size = buffer_size
        super().__init__()
        self.dp = dataset
        self._buffer = []
        self.key_func = key_func
        self.reverse = reverse

    def __iter__(self):
        for elem in self.dp:
            self._buffer.append(elem)
            if len(self._buffer) >= self.buffer_size:
                self._buffer.sort(key=self.key_func, reverse=self.reverse)
                for x in self._buffer:
                    yield x
                del self._buffer
                self._buffer = []
        # The sample left over
        self._buffer.sort(key=self.key_func, reverse=self.reverse)
        for x in self._buffer:
            yield x
        del self._buffer
        self._buffer = []


@functional_datapipe("dynamic_batch")
class DynamicBatchDataPipe(IterDataPipe):

    def __init__(self, dataset: IterDataPipe, window_class,
                 wrapper_class) -> None:
        _check_unpickable_fn(window_class)
        _check_unpickable_fn(wrapper_class)
        super().__init__()
        self.dp = dataset
        assert window_class is not None
        assert wrapper_class is not None
        self.window_class = window_class
        self._buffer = []
        self._wrappr_class = wrapper_class

    def __iter__(self):
        for elem in self.dp:
            if not self.window_class(elem, len(self._buffer)):
                self._buffer.append(elem)
            else:
                if len(self._buffer) > 0:
                    yield self._wrappr_class(self._buffer)
                del self._buffer
                self._buffer = [elem]
        if len(self._buffer) > 0:
            yield self._wrappr_class(self._buffer)
        del self._buffer
        self._buffer = []


@functional_datapipe("prefetch")
class PrefetchDataPipe(IterDataPipe):
    """Performs prefetching"""

    def __init__(
        self,
        dataset: IterDataPipe,
        buffer_size: int = 500,
    ):
        # TODO(Mddct): support multiprocessing pool with shared-memory to
        #   prefetch
        super().__init__()
        self.dp = dataset
        self._iter = None
        self._prefetch_buffer_size = buffer_size
        self._buffer = None
        if self._prefetch_buffer_size > 0:
            self._buffer = collections.deque(maxlen=self._prefetch_buffer_size)

    def __iter__(self):
        if self._prefetch_buffer_size > 0:
            if self._iter is None:
                self._iter = iter(self.dp)
            assert self._buffer is not None

            while True:
                if len(self._buffer) <= self._prefetch_buffer_size // 2:
                    while len(self._buffer) < self._prefetch_buffer_size:
                        try:
                            self._buffer.append(next(self._iter))
                        except StopIteration:
                            if len(self._buffer) != 0:
                                while len(self._buffer) > 0:
                                    yield self._buffer.popleft()
                            self._iter = None
                            return
                while len(self._buffer) > self._prefetch_buffer_size // 2:
                    elem = self._buffer.popleft()
                    yield elem

        else:
            yield from self.dp


@functional_datapipe("shard")
class ShardDataPipe(ShardingFilterIterDataPipe):

    def __init__(self, dataset: IterDataPipe, partition: bool = False):
        super().__init__(dataset, None)
        self.partition = partition
        self.dp = dataset

    def apply_sharding(self, num_of_instances: int, instance_id: int,
                       sharding_group: SHARDING_PRIORITIES):
        if self.partition:
            return super().apply_sharding(num_of_instances, instance_id,
                                          sharding_group)
        else:
            # We can not handle uneven data for CV on DDP, so we don't
            # sample data by rank, that means every GPU gets the same
            # and all the CV data
            info = torch.utils.data.get_worker_info()
            if info is None:
                self.num_of_instances = 1
                self.instance_id = 0
            else:
                n_workers_per_device = info.num_workers
                self.num_of_instances = n_workers_per_device
                self.instance_id = info.id


class TextLineDataPipe(IterDataPipe):
    """ Streamming Text line
    """

    def __init__(self, filenames, mode='r'):
        super().__init__()
        _dp = datapipes.iter.FileLister(filenames)
        _dp = datapipes.iter.FileOpener(_dp, mode=mode)
        self.dp = _dp

    def __iter__(self):
        for fname, stream in self.dp:
            for line in stream:
                line = line.strip('\n')
                yield {"file_name": fname, "line": line}
            stream.close()


@functional_datapipe("tar_file_and_group")
class TarsDataPipe(IterDataPipe):
    """ Decode wenet's tar , yield {'txt': "...", "raw": "..."}
    """

    def __init__(self, dataset: IterDataPipe) -> None:
        super().__init__()
        self.dp = dataset

    def __iter__(self):
        from wenet.dataset.processor import AUDIO_FORMAT_SETS
        for sample in self.dp:
            assert 'file_name' in sample
            assert 'line' in sample
            assert 'stream' in sample
            try:
                with tarfile.open(fileobj=sample['stream'],
                                  mode="r:*") as stream:
                    prev_prefix = None
                    example = {
                        'file_name': sample['file_name'],
                        'tar_file_name': sample['line']
                    }
                    valid = True
                    for tarinfo in stream:
                        name = tarinfo.name
                        pos = name.rfind('.')
                        assert pos > 0
                        prefix, postfix = name[:pos], name[pos + 1:]
                        if prev_prefix is not None and prefix != prev_prefix:
                            example['key'] = prev_prefix
                            if valid:
                                yield example
                            example = {
                                'file_name': sample['file_name'],
                                'tar_file_name': sample['line']
                            }
                            valid = True
                        with stream.extractfile(tarinfo) as file_obj:
                            try:
                                if postfix == 'txt':
                                    example['txt'] = file_obj.read().decode(
                                        'utf8').strip()
                                elif postfix in AUDIO_FORMAT_SETS:
                                    example['wav'] = file_obj.read()
                                else:
                                    example[postfix] = file_obj.read()
                            except Exception as ex:
                                valid = False
                                logging.warning(
                                    'error to parse {}'.format(name))
                            prev_prefix = prefix
                    if prev_prefix is not None:
                        example['key'] = prev_prefix
                        yield example
            except Exception as ex:
                msg = 'In tar_file_and_group: {} when processing {}'.format(
                    ex, sample['line'])
                logging.warning(msg)
            finally:
                if 'process' in sample:
                    sample['process'].communicate()
                sample['stream'].close()


class WenetRawDatasetSource(IterDataPipe):

    def __init__(self,
                 filenames: str,
                 prefetch: int = 500,
                 partition=True) -> None:
        super().__init__()
        self.dp = TextLineDataPipe(filenames).prefetch(prefetch).shard(
            partition)

    def __iter__(self):
        for d in self.dp:
            yield d


class WenetTarShardDatasetSource(IterDataPipe):

    def __init__(self,
                 filenames: str,
                 prefetch: int = 500,
                 partition: bool = False) -> None:
        super().__init__()
        self.dp = TextLineDataPipe(filenames).shard(
            partition).map_ignore_error(
                parse_url).tar_file_and_group().prefetch(prefetch)

    def __iter__(self):
        for d in self.dp:
            yield d



from gxl_ai_utils.utils import utils_file

input_jsonl_path = "/mnt/disk1/yhdai/data_zipformer/batch01/aishell1/fbank/modified_cuts_train.jsonl"
data_list = utils_file.load_dict_list_from_jsonl(input_jsonl_path)
duration = 0
for data in utils_file.tqdm(data_list, total=len(data_list)):
    duration += data["duration"]
print(duration)


import argparse
import codecs
import os

import jieba
import tqdm
from nltk.translate.bleu_score import SmoothingFunction, sentence_bleu


def calculate_bleu(candidate_text: str, reference_texts: list) -> float:
    """
    计算中文对话任务的 BLEU 分数。

    参数:
        candidate_text (str): 生成的文本。
        reference_texts (list): 参考文本列表（可以有多个参考句子）。

    返回:
        float: BLEU 分数。
    """
    # 对生成文本和参考文本进行分词
    candidate_tokens = list(jieba.cut(candidate_text))
    reference_tokens = [list(jieba.cut(ref)) for ref in reference_texts]

    # 使用 nltk 的 sentence_bleu 计算 BLEU 分数
    smoothing = SmoothingFunction().method1  # 避免分数为零
    bleu_score = sentence_bleu(reference_tokens, candidate_tokens,
                               weights=(0.25, 0.25, 0.25, 0.25),  # 1-gram 到 4-gram 的权重
                               smoothing_function=smoothing)

    return bleu_score

def load_dict_from_scp(label_scp_file: str, silence: bool = False) -> dict:
    """
    得到scp文件的内容,要求key value以空格或者tab分割， 第一个为key,剩下的都是value。
    :param label_scp_file:
    :return:
    """
    res = {}
    with codecs.open(label_scp_file, 'r', encoding='utf-8') as f:
        try:
            lines = f.readlines()
        except Exception as e:
            print(e)
            return {}
        for line in lines:
            line = line.strip()
            items = line.split()
            if len(items) < 2:
                if not silence:
                    print(
                        'load_dict_from_scp;warning_gxl:, this row not conform to the regulation of scp(key content) and skip it:',
                        line)
                continue
            elif len(items) == 2:
                res[items[0].strip()] = items[1].strip()
            else:
                # logging_print(
                #     'warning_gxl:, this row not conform to the regulation of'
                #     ' scp(key content) and no skip it,第一个为key,剩下的都是value:',
                #     line)
                res[items[0].strip()] = (' '.join(items[1:])).strip()
    total_len = len(res)
    print("load_dict_from_scp()_数据总条数为:", total_len)
    return res

# 得到命令行参数, 使用parser.add_argument()方法
parser = argparse.ArgumentParser()
parser.add_argument('--ref_file', type=str, required=True, help='reference file')
parser.add_argument('--hyp_file', type=str, required=True, help='hypothesis file')
parser.add_argument('--output_file', type=str, required=True, help='output file')
args = parser.parse_args()

# 计算BLEU得分
ref_dict = load_dict_from_scp(args.ref_file)
hyp_dict = load_dict_from_scp(args.hyp_file)
# os.makedirs(os.path.dirname(args.output_file), exist_ok=True)
res_dict = {}
total_bleu_score = 0
for key, value in tqdm.tqdm(hyp_dict.items(), total=len(hyp_dict)):
    if key in ref_dict:
        ref_str = ref_dict[key]
        hyp_str = value
        bleu_score = calculate_bleu(hyp_str, [ref_str])
        res_dict[key] = bleu_score
        total_bleu_score += bleu_score
    else:
        print(f"key {key} not in ref_dict")
        continue
# 保存结果到文件
def write_dict_to_scp(dic: dict, scp_file_path: str):
    print("开始write_dict_to_scp()，数据总条数为:", len(dic))
    os.makedirs(os.path.dirname(scp_file_path), exist_ok=True)
    with codecs.open(scp_file_path, 'w', encoding='utf-8') as f:
        for k, v in dic.items():
            f.write(f"{k} {v}\n")
avg_bleu_score = total_bleu_score / len(res_dict)
print(f"平均BLEU得分为:{avg_bleu_score}")
res_dict['avg_bleu_score'] = avg_bleu_score
write_dict_to_scp(res_dict, args.output_file)


import subprocess

import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers.generation import GenerationConfig
import torch
import os
from gxl_ai_utils.utils import utils_file

torch.manual_seed(1234)
os.environ["CUDA_VISIBLE_DEVICES"] = "1"


# 打开bf16精度，A100、H100、RTX3060、RTX3070等显卡建议启用以节省显存
# model = AutoModelForCausalLM.from_pretrained("Qwen/Qwen-Audio", device_map="auto", trust_remote_code=True, bf16=True).eval()
# 打开fp16精度，V100、P100、T4等显卡建议启用以节省显存
# model = AutoModelForCausalLM.from_pretrained("Qwen/Qwen-Audio", device_map="auto", trust_remote_code=True, fp16=True).eval()
# 使用CPU进行推理，需要约32GB内存
# model = AutoModelForCausalLM.from_pretrained("Qwen/Qwen-Audio", device_map="cpu", trust_remote_code=True).eval()
# 默认gpu进行推理，需要约24GB显存
# model = AutoModelForCausalLM.from_pretrained("Qwen/Qwen-Audio", trust_remote_code=True).eval()
# 可指定不同的生成长度、top_p等相关超参（transformers 4.32.0及以上无需执行此操作）
# model.generation_config = GenerationConfig.from_pretrained("Qwen/Qwen-Audio", trust_remote_code=True)

def compute_wer(true_text_path, hyp_text_path, output_dir):
    os.makedirs(output_dir, exist_ok=True)
    command = "python compute-wer.py --char=1 --v=1 \
          {} {} > {}/wer".format(true_text_path, hyp_text_path, output_dir)
    # 执行命令
    subprocess.run(command, shell=True)


def model_forward(model, tokenizer, wav_path):
    # audio_url = "/home/work_nfs/common/data/data_aishell/wav/test/S0764/BAC009S0764W0124.wav"
    sp_prompt = "<|startoftranscription|><|zh|><|transcribe|><|zh|><|notimestamps|><|wo_itn|>"
    query = f"<audio>{wav_path}</audio>{sp_prompt}"
    audio_info = tokenizer.process_audio(query)
    inputs = tokenizer(query, return_tensors='pt', audio_info=audio_info)
    inputs = inputs.to(model.device)
    pred = model.generate(**inputs, audio_info=audio_info)
    response = tokenizer.decode(pred.cpu()[0], skip_special_tokens=True, audio_info=audio_info)
    response = response.split('<|startoftranscription|>')
    # key = utils_file.get_file_pure_name_from_path(response[0])
    res = response[1]
    return res


def recognize():
    """"""
    tokenizer = AutoTokenizer.from_pretrained(
        "/home/work_nfs8/xlgeng/.cache/transformers/models--Qwen--Qwen-Audio/snapshots/cbb9c956a096a5d5420d6c4927fef30d00eff144",
        trust_remote_code=True)
    model = AutoModelForCausalLM.from_pretrained(
        "/home/work_nfs8/xlgeng/.cache/transformers/models--Qwen--Qwen-Audio/snapshots/cbb9c956a096a5d5420d6c4927fef30d00eff144",
        device_map="cuda", trust_remote_code=True).eval()
    output_dir = './res_out_kespeech'
    os.makedirs(output_dir, exist_ok=True)
    data_dir = "/home/work_nfs8/xlgeng/data/scp_test"
    data_dir = "/home/work_nfs8/xlgeng/new_workspace/wenet_gxl_salmonn/examples/aishell/salmonn/data_list/kespeech"
    test_sets = ['test_net', 'test_meeting', 'speechio_0', 'speechio_1', 'speechio_2', 'speechio_3', 'speechio_4']
    test_sets = ["Beijing", "Jiang-Huai", "Jiao-Liao", "Ji-Lu", "Lan-Yin", "Northeastern", "Southwestern", "Zhongyuan"]
    for test_set in test_sets:
        utils_file.logging_print(f"耿雪龙: 开始测试test_set: {test_set}")
        res_dict = {}
        test_set_dir = os.path.join(data_dir, test_set, 'train')
        wav_scp_path = os.path.join(test_set_dir, 'wav.scp')
        text_scp_path = os.path.join(test_set_dir, 'text')
        wav_line_list = utils_file.load_list_file_clean(wav_scp_path)
        for wav_line in wav_line_list:
            items = wav_line.split()
            if len(items) != 2:
                continue
            key = items[0]
            wav_path = items[1]
            text_res = model_forward(model, tokenizer, wav_path)
            utils_file.logging_print(f'{key}: {text_res}')
            res_dict[key] = text_res
        output_dir_i = os.path.join(output_dir, test_set)
        utils_file.makedir_sil(output_dir_i)
        utils_file.copy_file2(text_scp_path, output_dir_i)
        output_path = os.path.join(output_dir_i, f"{test_set}_text")
        utils_file.write_dict_to_scp(res_dict, output_path)
        utils_file.compute_wer(text_scp_path, output_path, output_dir_i)


def handle_log():
    input_log_path = "/home/work_nfs8/xlgeng/new_workspace/gxl_ai_utils/eggs/cats_and_dogs/prepare_data_for_salmonn/output/run_res/gxl.nohup"
    lines = utils_file.load_list_file_clean(input_log_path)
    res_dict = {}
    key_word = 'INFO TEST_NET_Y0000'
    for lines in tqdm.tqdm(lines, total=len(lines)):
        if key_word not in lines:
            continue
        items = lines.split()
        key = items[3]
        key = key.split(":")[0]
        val = " ".join(items[4:])
        res_dict[key] = val
    output_path = "./res_out/text_net/text_hyp"
    utils_file.makedir_for_file(output_path)
    utils_file.write_dict_to_scp(res_dict, output_path)
    true_path = "/home/work_nfs8/xlgeng/data/scp_test/test_net/text"
    utils_file.compute_wer(true_path, output_path, "./res_out/text_net")


if __name__ == '__main__':
    # true_path = "/home/work_nfs8/xlgeng/data/scp_test/aishell/text"
    # hyp_path = "/home/work_nfs8/xlgeng/new_workspace/wenet_gxl_salmonn/examples/aishell/salmonn/exp/salmonn_v15/2_epoch_new/test_step_16422/aishell/text"
    # output_dir = "./res_out"
    # utils_file.compute_wer(true_path, hyp_path, output_dir)
    # compute_wer(true_path,hyp_path,output_dir)
    # handle_log()
    recognize()



import os
import sys
sys.path.insert(0,'../../../../')
from gxl_ai_utils.utils import utils_file


input_dir = "/home/work_nfs15/asr_data/data/LibriSpeech"
output_dir = "/home/work_nfs15/asr_data/data/LibriSpeech/LibriSpeech_shard"
parts = ["train-clean-100",  "train-other-500", "train-clean-360"]
for part in parts:
    utils_file.logging_info('开始处理：', part)
    input_dir_part = os.path.join(input_dir, part)
    output_dir_part = os.path.join(output_dir, part)
    utils_file.makedir_sil(output_dir_part)
    scp_wav_path = os.path.join(output_dir_part, "wav.scp")
    text_path = os.path.join(output_dir_part, "text")
    if part == 'train-other-500':
        utils_file.logging_info('获取wav.scp')
        file_path_list = utils_file.do_get_list_for_wav_dir(input_dir_part, None, '.flac', True)
        wav_dict = utils_file.do_convert_file_list_to_dict(file_path_list)
        utils_file.write_dict_to_scp(wav_dict, scp_wav_path)
        text_all_dict = utils_file.load_dict_from_scp("/home/work_nfs15/asr_data/data/LibriSpeech/LibriSpeech_shard/text")
        text_dict = {k: text_all_dict[k] for k in wav_dict.keys()}
        utils_file.write_dict_to_scp(text_dict, text_path)
        utils_file.logging_info('得到scp完毕， 开始通过scp得到shard包')
        utils_file.do_make_shard_file(scp_wav_path, text_path, os.path.join(output_dir_part, 'shards'),
                                      num_utt_per_shard=1000, num_threads=50)
        continue

    else:
        pass

    utils_file.logging_info('获取wav.scp')
    list_path = os.path.join(input_dir_part, "list")
    file_path_list = utils_file.do_replace_str_to_file_and_return_list(
        source_str="/home/disk2/librispeech/LibriSpeech",
        target_str="/home/work_nfs15/asr_data/data/LibriSpeech",
        input_file=list_path
    )
    wav_dict = utils_file.do_convert_file_list_to_dict(file_path_list)
    utils_file.write_dict_to_scp(wav_dict, scp_wav_path)

    source_text_path = os.path.join(input_dir_part, f"{part}.trans")
    utils_file.copy_file(source_text_path, text_path)
    utils_file.logging_info('得到scp完毕， 开始通过scp得到shard包')
    utils_file.do_make_shard_file(scp_wav_path, text_path, os.path.join(output_dir_part, 'shards'), num_utt_per_shard=1000, num_threads=50)






# Copyright (c) 2023 ASLP@NWPU (authors: He Wang, Fan Yu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License. Modified from
# FunASR(https://github.com/alibaba-damo-academy/FunASR)

from typing import Optional

import torch
from torch import nn
from torchaudio.compliance.kaldi import Tuple
from ..utils.mask import make_pad_mask


class Predictor(nn.Module):

    def __init__(self,
                 idim,
                 l_order,
                 r_order,
                 threshold=1.0,
                 dropout=0.1,
                 smooth_factor=1.0,
                 noise_threshold=0,
                 tail_threshold=0.45,
                 residual=True,
                 cnn_groups=0):
        super().__init__()

        self.pad = nn.ConstantPad1d((l_order, r_order), 0.0)
        self.cif_conv1d = nn.Conv1d(
            idim,
            idim,
            l_order + r_order + 1,
            groups=idim if cnn_groups == 0 else cnn_groups)
        self.cif_output = nn.Linear(idim, 1)
        self.dropout = torch.nn.Dropout(p=dropout)
        self.threshold = threshold
        self.smooth_factor = smooth_factor
        self.noise_threshold = noise_threshold
        self.tail_threshold = tail_threshold
        self.residual = residual

    def forward(self,
                hidden,
                target_label: Optional[torch.Tensor] = None,
                mask: torch.Tensor = torch.tensor(0),
                ignore_id: int = -1,
                mask_chunk_predictor: Optional[torch.Tensor] = None,
                target_label_length: Optional[torch.Tensor] = None):
        h = hidden
        context = h.transpose(1, 2)
        queries = self.pad(context)
        memory = self.cif_conv1d(queries)
        if self.residual:
            output = memory + context
        else:
            output = memory
        output = self.dropout(output)
        output = output.transpose(1, 2)
        output = torch.relu(output)
        output = self.cif_output(output)
        alphas = torch.sigmoid(output)
        alphas = torch.nn.functional.relu(alphas * self.smooth_factor -
                                          self.noise_threshold)
        if mask is not None:
            mask = mask.transpose(-1, -2)
            alphas = alphas * mask
        if mask_chunk_predictor is not None:
            alphas = alphas * mask_chunk_predictor
        alphas = alphas.squeeze(-1)
        mask = mask.squeeze(-1)
        if target_label_length is not None:
            target_length = target_label_length
        elif target_label is not None:
            target_length = (target_label != ignore_id).float().sum(-1)
        else:
            target_length = None
        token_num = alphas.sum(-1)
        if target_length is not None:
            alphas *= (target_length / token_num)[:, None] \
                .repeat(1, alphas.size(1))
        elif self.tail_threshold > 0.0:
            hidden, alphas, token_num = self.tail_process_fn(hidden,
                                                             alphas,
                                                             token_num,
                                                             mask=mask)
        acoustic_embeds, cif_peak = cif(hidden, alphas, self.threshold)

        if target_length is None and self.tail_threshold > 0.0:
            token_num_int = torch.max(token_num).type(torch.int32).item()
            acoustic_embeds = acoustic_embeds[:, :token_num_int, :]

        return acoustic_embeds, token_num, alphas, cif_peak

    def tail_process_fn(
        self,
        hidden: torch.Tensor,
        alphas: torch.Tensor,
        token_num: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        b, _, d = hidden.size()
        if mask is not None:
            zeros_t = torch.zeros((b, 1),
                                  dtype=torch.float32,
                                  device=alphas.device)
            mask = mask.to(zeros_t.dtype)
            ones_t = torch.ones_like(zeros_t)
            mask_1 = torch.cat([mask, zeros_t], dim=1)
            mask_2 = torch.cat([ones_t, mask], dim=1)
            mask = mask_2 - mask_1
            tail_threshold = mask * self.tail_threshold
            alphas = torch.cat([alphas, zeros_t], dim=1)
            alphas = torch.add(alphas, tail_threshold)
        else:
            tail_threshold_tensor = torch.tensor([self.tail_threshold],
                                                 dtype=alphas.dtype).to(
                                                     alphas.device)
            tail_threshold_tensor = torch.reshape(tail_threshold_tensor,
                                                  (1, 1))
            alphas = torch.cat([alphas, tail_threshold_tensor], dim=1)
        zeros = torch.zeros((b, 1, d), dtype=hidden.dtype).to(hidden.device)
        hidden = torch.cat([hidden, zeros], dim=1)
        token_num = alphas.sum(dim=-1)
        token_num_floor = torch.floor(token_num)

        return hidden, alphas, token_num_floor

    def gen_frame_alignments(self,
                             alphas: torch.Tensor = None,
                             encoder_sequence_length: torch.Tensor = None):
        batch_size, maximum_length = alphas.size()
        int_type = torch.int32

        is_training = self.training
        if is_training:
            token_num = torch.round(torch.sum(alphas, dim=1)).type(int_type)
        else:
            token_num = torch.floor(torch.sum(alphas, dim=1)).type(int_type)

        max_token_num = torch.max(token_num).item()

        alphas_cumsum = torch.cumsum(alphas, dim=1)
        alphas_cumsum = torch.floor(alphas_cumsum).type(int_type)
        alphas_cumsum = alphas_cumsum[:, None, :].repeat(1, max_token_num, 1)

        index = torch.ones([batch_size, max_token_num], dtype=int_type)
        index = torch.cumsum(index, dim=1)
        index = index[:, :,
                      None].repeat(1, 1,
                                   maximum_length).to(alphas_cumsum.device)

        index_div = torch.floor(torch.true_divide(alphas_cumsum,
                                                  index)).type(int_type)
        index_div_bool_zeros = index_div.eq(0)
        index_div_bool_zeros_count = torch.sum(index_div_bool_zeros,
                                               dim=-1) + 1
        index_div_bool_zeros_count = torch.clamp(index_div_bool_zeros_count, 0,
                                                 encoder_sequence_length.max())
        token_num_mask = (~make_pad_mask(token_num, max_len=max_token_num)).to(
            token_num.device)
        index_div_bool_zeros_count *= token_num_mask

        index_div_bool_zeros_count_tile = \
            index_div_bool_zeros_count[:, :, None].repeat(1, 1, maximum_length)
        ones = torch.ones_like(index_div_bool_zeros_count_tile)
        zeros = torch.zeros_like(index_div_bool_zeros_count_tile)
        ones = torch.cumsum(ones, dim=2)
        cond = index_div_bool_zeros_count_tile == ones
        index_div_bool_zeros_count_tile = torch.where(cond, zeros, ones)

        index_div_bool_zeros_count_tile_bool = index_div_bool_zeros_count_tile \
            .type(torch.bool)
        index_div_bool_zeros_count_tile = \
            1 - index_div_bool_zeros_count_tile_bool.type(int_type)
        index_div_bool_zeros_count_tile_out = torch.sum(
            index_div_bool_zeros_count_tile, dim=1)
        index_div_bool_zeros_count_tile_out = \
            index_div_bool_zeros_count_tile_out.type(int_type)
        predictor_mask = (~make_pad_mask(encoder_sequence_length,
                                         max_len=encoder_sequence_length
                                         .max())).type(int_type)\
            .to(encoder_sequence_length.device)
        index_div_bool_zeros_count_tile_out = \
            index_div_bool_zeros_count_tile_out * predictor_mask

        predictor_alignments = index_div_bool_zeros_count_tile_out
        predictor_alignments_length = predictor_alignments.sum(-1).type(
            encoder_sequence_length.dtype)
        return predictor_alignments.detach(), \
            predictor_alignments_length.detach()


class MAELoss(nn.Module):

    def __init__(self, normalize_length=False):
        super(MAELoss, self).__init__()
        self.normalize_length = normalize_length
        self.criterion = torch.nn.L1Loss(reduction='sum')

    def forward(self, token_length, pre_token_length):
        loss_token_normalizer = token_length.size(0)
        if self.normalize_length:
            loss_token_normalizer = token_length.sum().type(torch.float32)
        loss = self.criterion(token_length, pre_token_length)
        loss = loss / loss_token_normalizer
        return loss


def cif(hidden: torch.Tensor, alphas: torch.Tensor, threshold: float):
    batch_size, len_time, hidden_size = hidden.size()

    # loop varss
    integrate = torch.zeros([batch_size], device=hidden.device)
    frame = torch.zeros([batch_size, hidden_size], device=hidden.device)
    # intermediate vars along time
    list_fires = []
    list_frames = []

    for t in range(len_time):
        alpha = alphas[:, t]
        distribution_completion = torch.ones([batch_size],
                                             device=hidden.device) - integrate

        integrate += alpha
        list_fires.append(integrate)

        fire_place = integrate >= threshold
        integrate = torch.where(
            fire_place,
            integrate - torch.ones([batch_size], device=hidden.device),
            integrate)
        cur = torch.where(fire_place, distribution_completion, alpha)
        remainds = alpha - cur

        frame += cur[:, None] * hidden[:, t, :]
        list_frames.append(frame)
        frame = torch.where(fire_place[:, None].repeat(1, hidden_size),
                            remainds[:, None] * hidden[:, t, :], frame)

    fires = torch.stack(list_fires, 1)
    frames = torch.stack(list_frames, 1)
    list_ls = []
    len_labels = torch.round(alphas.sum(-1)).int()
    max_label_len = len_labels.max()
    for b in range(batch_size):
        fire = fires[b, :]
        l = torch.index_select(frames[b, :, :], 0,
                               torch.nonzero(fire >= threshold).squeeze())
        pad_l = torch.zeros([int(max_label_len - l.size(0)), hidden_size],
                            device=hidden.device)
        list_ls.append(torch.cat([l, pad_l], 0))
    return torch.stack(list_ls, 0), fires



import os

from gxl_ai_utils.utils import utils_file

input_train_list_path = "/home/work_nfs14/yhdai/data_shard/chattts/01/shards_list.txt"
input_cv_list_path = "/home/work_nfs14/yhdai/data_shard/chattts/cv/shards_list.txt"
output_dir = "/home/work_nfs15/xlgeng/data/data_shards/chat_wav_data"
# train
output_dir_train = os.path.join(output_dir, "train")
utils_file.makedir_sil(output_dir_train)
new_list_path = os.path.join(output_dir_train, "shards_list.txt")
path_list = utils_file.load_list_file_clean(input_train_list_path)
new_path_list = []
for path in path_list:
    new_path_i = utils_file.do_replace_dir(path, output_dir_train)
    utils_file.copy_file(path, new_path_i)
    new_path_list.append(new_path_i)
utils_file.write_list_to_file(new_path_list, new_list_path)

# cv
output_dir_train = os.path.join(output_dir, "cv")
utils_file.makedir_sil(output_dir_train)
new_list_path = os.path.join(output_dir_train, "shards_list.txt")
path_list = utils_file.load_list_file_clean(input_cv_list_path)
new_path_list = []
for path in path_list:
    new_path_i = utils_file.do_replace_dir(path, output_dir_train)
    utils_file.copy_file(path, new_path_i)
    new_path_list.append(new_path_i)
utils_file.write_list_to_file(new_path_list, new_list_path)








# Copyright    2022  Xiaomi Corp.        (authors: Wei Kang,
#                                                  Zengwei Yao)
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import List, Optional, Tuple

import k2
import torch
from beam_search import Hypothesis, HypothesisList

from icefall.utils import AttributeDict


class DecodeStream(object):
    def __init__(
        self,
        params: AttributeDict,
        cut_id: str,
        initial_states: List[torch.Tensor],
        decoding_graph: Optional[k2.Fsa] = None,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        """
        Args:
          initial_states:
            Initial decode states of the model, e.g. the return value of
            `get_init_state` in conformer.py
          decoding_graph:
            Decoding graph used for decoding, may be a TrivialGraph or a HLG.
            Used only when decoding_method is fast_beam_search.
          device:
            The device to run this stream.
        """
        if params.decoding_method == "fast_beam_search":
            assert decoding_graph is not None
            assert device == decoding_graph.device

        self.params = params
        self.cut_id = cut_id
        self.LOG_EPS = math.log(1e-10)

        self.states = initial_states

        # It contains a 2-D tensors representing the feature frames.
        self.features: torch.Tensor = None

        self.num_frames: int = 0
        # how many frames have been processed. (before subsampling).
        # we only modify this value in `func:get_feature_frames`.
        self.num_processed_frames: int = 0

        self._done: bool = False

        # The transcript of current utterance.
        self.ground_truth: str = ""

        # The decoding result (partial or final) of current utterance.
        self.hyp: List = []

        # how many frames have been processed, at encoder output
        self.done_frames: int = 0

        # The encoder_embed subsample features (T - 7) // 2
        # The ConvNeXt module needs (7 - 1) // 2 = 3 frames of right padding after subsampling
        self.pad_length = 7 + 2 * 3

        if params.decoding_method == "greedy_search":
            self.hyp = [-1] * (params.context_size - 1) + [params.blank_id]
        elif params.decoding_method == "modified_beam_search":
            self.hyps = HypothesisList()
            self.hyps.add(
                Hypothesis(
                    ys=[-1] * (params.context_size - 1) + [params.blank_id],
                    log_prob=torch.zeros(1, dtype=torch.float32, device=device),
                )
            )
        elif params.decoding_method == "fast_beam_search":
            # The rnnt_decoding_stream for fast_beam_search.
            self.rnnt_decoding_stream: k2.RnntDecodingStream = k2.RnntDecodingStream(
                decoding_graph
            )
        else:
            raise ValueError(f"Unsupported decoding method: {params.decoding_method}")

    @property
    def done(self) -> bool:
        """Return True if all the features are processed."""
        return self._done

    @property
    def id(self) -> str:
        return self.cut_id

    def set_features(
        self,
        features: torch.Tensor,
        tail_pad_len: int = 0,
    ) -> None:
        """Set features tensor of current utterance."""
        assert features.dim() == 2, features.dim()
        self.features = torch.nn.functional.pad(
            features,
            (0, 0, 0, self.pad_length + tail_pad_len),
            mode="constant",
            value=self.LOG_EPS,
        )
        self.num_frames = self.features.size(0)

    def get_feature_frames(self, chunk_size: int) -> Tuple[torch.Tensor, int]:
        """Consume chunk_size frames of features"""
        chunk_length = chunk_size + self.pad_length

        ret_length = min(self.num_frames - self.num_processed_frames, chunk_length)

        ret_features = self.features[
            self.num_processed_frames : self.num_processed_frames + ret_length  # noqa
        ]

        self.num_processed_frames += chunk_size
        if self.num_processed_frames >= self.num_frames:
            self._done = True

        return ret_features, ret_length

    def decoding_result(self) -> List[int]:
        """Obtain current decoding result."""
        if self.params.decoding_method == "greedy_search":
            return self.hyp[self.params.context_size :]  # noqa
        elif self.params.decoding_method == "modified_beam_search":
            best_hyp = self.hyps.get_most_probable(length_norm=True)
            return best_hyp.ys[self.params.context_size :]  # noqa
        else:
            assert self.params.decoding_method == "fast_beam_search"
            return self.hyp
